# -*- coding: utf-8 -*-
# coding: utf-8
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import zip, next, range, str

from ....pipeline import engine as pe
from ....interfaces import utility as niu
from ....interfaces import fsl
from ....interfaces import ants


def cleanup_edge_pipeline(name='Cleanup'):
    """
    Perform some de-spiking filtering to clean up the edge of the fieldmap
    (copied from fsl_prepare_fieldmap)
    """
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_file', 'in_mask']), name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file']), name='outputnode')

    fugue = pe.Node(
        fsl.FUGUE(
            save_fmap=True, despike_2dfilter=True, despike_threshold=2.1),
        name='Despike')
    erode = pe.Node(
        fsl.maths.MathsCommand(nan2zeros=True, args='-kernel 2D -ero'),
        name='MskErode')
    newmsk = pe.Node(
        fsl.MultiImageMaths(op_string='-sub %s -thr 0.5 -bin'), name='NewMask')
    applymsk = pe.Node(fsl.ApplyMask(nan2zeros=True), name='ApplyMask')
    join = pe.Node(niu.Merge(2), name='Merge')
    addedge = pe.Node(
        fsl.MultiImageMaths(op_string='-mas %s -add %s'), name='AddEdge')

    wf = pe.Workflow(name=name)
    wf.connect([(inputnode, fugue, [
        ('in_file', 'fmap_in_file'), ('in_mask', 'mask_file')
    ]), (inputnode, erode, [('in_mask', 'in_file')]), (inputnode, newmsk, [
        ('in_mask', 'in_file')
    ]), (erode, newmsk, [('out_file', 'operand_files')]), (fugue, applymsk, [
        ('fmap_out_file', 'in_file')
    ]), (newmsk, applymsk,
         [('out_file', 'mask_file')]), (erode, join, [('out_file', 'in1')]),
                (applymsk, join, [('out_file', 'in2')]), (inputnode, addedge, [
                    ('in_file', 'in_file')
                ]), (join, addedge, [('out', 'operand_files')]),
                (addedge, outputnode, [('out_file', 'out_file')])])
    return wf


def vsm2warp(name='Shiftmap2Warping'):
    """
    Converts a voxel shift map (vsm) to a displacements field (warp).
    """
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['in_vsm', 'in_ref', 'scaling', 'enc_dir']),
        name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_warp']), name='outputnode')
    fixhdr = pe.Node(
        niu.Function(
            input_names=['in_file', 'in_file_hdr'],
            output_names=['out_file'],
            function=copy_hdr),
        name='Fix_hdr')
    vsm = pe.Node(fsl.maths.BinaryMaths(operation='mul'), name='ScaleField')
    vsm2dfm = pe.Node(
        fsl.ConvertWarp(relwarp=True, out_relwarp=True), name='vsm2dfm')

    wf = pe.Workflow(name=name)
    wf.connect([(inputnode, fixhdr, [('in_vsm', 'in_file'), ('in_ref',
                                                             'in_file_hdr')]),
                (inputnode, vsm,
                 [('scaling', 'operand_value')]), (fixhdr, vsm, [('out_file',
                                                                  'in_file')]),
                (vsm, vsm2dfm,
                 [('out_file', 'shift_in_file')]), (inputnode, vsm2dfm, [
                     ('in_ref', 'reference'), ('enc_dir', 'shift_direction')
                 ]), (vsm2dfm, outputnode, [('out_file', 'out_warp')])])
    return wf


def dwi_flirt(name='DWICoregistration', excl_nodiff=False, flirt_param={}):
    """
    Generates a workflow for linear registration of dwi volumes
    """
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['reference', 'in_file', 'ref_mask', 'in_xfms', 'in_bval']),
        name='inputnode')

    initmat = pe.Node(
        niu.Function(
            input_names=['in_bval', 'in_xfms', 'excl_nodiff'],
            output_names=['init_xfms'],
            function=_checkinitxfm),
        name='InitXforms')
    initmat.inputs.excl_nodiff = excl_nodiff
    dilate = pe.Node(
        fsl.maths.MathsCommand(nan2zeros=True, args='-kernel sphere 5 -dilM'),
        name='MskDilate')
    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    n4 = pe.Node(ants.N4BiasFieldCorrection(dimension=3), name='Bias')
    enhb0 = pe.Node(
        niu.Function(
            input_names=['in_file', 'in_mask', 'clip_limit'],
            output_names=['out_file'],
            function=enhance),
        name='B0Equalize')
    enhb0.inputs.clip_limit = 0.015
    enhdw = pe.MapNode(
        niu.Function(
            input_names=['in_file', 'in_mask'],
            output_names=['out_file'],
            function=enhance),
        name='DWEqualize',
        iterfield=['in_file'])
    flirt = pe.MapNode(
        fsl.FLIRT(**flirt_param),
        name='CoRegistration',
        iterfield=['in_file', 'in_matrix_file'])
    apply_xfms = pe.MapNode(
        fsl.ApplyXFM(
            apply_xfm=True,
            interp='spline',
            bgvalue=0),
        name='ApplyXFMs',
        iterfield=['in_file', 'in_matrix_file']
    )
    thres = pe.MapNode(
        fsl.Threshold(thresh=0.0),
        iterfield=['in_file'],
        name='RemoveNegative')
    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeDWIs')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file', 'out_xfms']),
        name='outputnode')
    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, split, [('in_file', 'in_file')]),
        (inputnode, dilate, [('ref_mask', 'in_file')]),
        (inputnode, enhb0, [('ref_mask', 'in_mask')]),
        (inputnode, initmat, [('in_xfms', 'in_xfms'),
                              ('in_bval', 'in_bval')]),
        (inputnode, n4, [('reference', 'input_image'),
                         ('ref_mask', 'mask_image')]),
        (dilate, flirt, [('out_file', 'ref_weight'),
                         ('out_file', 'in_weight')]),
        (n4, enhb0, [('output_image', 'in_file')]),
        (split, enhdw, [('out_files', 'in_file')]),
        (split, apply_xfms, [('out_files', 'in_file')]),
        (dilate, enhdw, [('out_file', 'in_mask')]),
        (enhb0, flirt, [('out_file', 'reference')]),
        (enhb0, apply_xfms, [('out_file', 'reference')]),
        (enhdw, flirt, [('out_file', 'in_file')]),
        (initmat, flirt, [('init_xfms', 'in_matrix_file')]),
        (flirt, apply_xfms, [('out_matrix_file', 'in_matrix_file')]),
        (apply_xfms, thres, [('out_file', 'in_file')]),
        (thres, merge, [('out_file', 'in_files')]),
        (merge, outputnode, [('merged_file', 'out_file')]),
        (flirt, outputnode, [('out_matrix_file', 'out_xfms')])
    ])
    return wf


def apply_all_corrections(name='UnwarpArtifacts'):
    """
    Combines two lists of linear transforms with the deformation field
    map obtained typically after the SDC process.
    Additionally, computes the corresponding bspline coefficients and
    the map of determinants of the jacobian.
    """

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_sdc', 'in_hmc', 'in_ecc', 'in_dwi']),
        name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=['out_file', 'out_warp', 'out_coeff', 'out_jacobian']),
        name='outputnode')
    warps = pe.MapNode(
        fsl.ConvertWarp(relwarp=True),
        iterfield=['premat', 'postmat'],
        name='ConvertWarp')

    selref = pe.Node(niu.Select(index=[0]), name='Reference')

    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    unwarp = pe.MapNode(
        fsl.ApplyWarp(),
        iterfield=['in_file', 'field_file'],
        name='UnwarpDWIs')

    coeffs = pe.MapNode(
        fsl.WarpUtils(out_format='spline'),
        iterfield=['in_file'],
        name='CoeffComp')
    jacobian = pe.MapNode(
        fsl.WarpUtils(write_jacobian=True),
        iterfield=['in_file'],
        name='JacobianComp')
    jacmult = pe.MapNode(
        fsl.MultiImageMaths(op_string='-mul %s'),
        iterfield=['in_file', 'operand_files'],
        name='ModulateDWIs')

    thres = pe.MapNode(
        fsl.Threshold(thresh=0.0),
        iterfield=['in_file'],
        name='RemoveNegative')
    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeDWIs')

    wf = pe.Workflow(name=name)
    wf.connect([(inputnode, warps, [
        ('in_sdc', 'warp1'), ('in_hmc', 'premat'), ('in_ecc', 'postmat'),
        ('in_dwi', 'reference')
    ]), (inputnode, split, [('in_dwi', 'in_file')]), (split, selref, [
        ('out_files', 'inlist')
    ]), (warps, unwarp, [('out_file', 'field_file')]), (split, unwarp, [
        ('out_files', 'in_file')
    ]), (selref, unwarp, [('out', 'ref_file')]), (selref, coeffs, [
        ('out', 'reference')
    ]), (warps, coeffs, [('out_file', 'in_file')]), (selref, jacobian, [
        ('out', 'reference')
    ]), (coeffs, jacobian, [('out_file', 'in_file')]), (unwarp, jacmult, [
        ('out_file', 'in_file')
    ]), (jacobian, jacmult, [('out_jacobian', 'operand_files')]),
                (jacmult, thres, [('out_file', 'in_file')]), (thres, merge, [
                    ('out_file', 'in_files')
                ]), (warps, outputnode, [('out_file', 'out_warp')]),
                (coeffs, outputnode,
                 [('out_file', 'out_coeff')]), (jacobian, outputnode, [
                     ('out_jacobian', 'out_jacobian')
                 ]), (merge, outputnode, [('merged_file', 'out_file')])])
    return wf


def extract_bval(in_dwi, in_bval, b=0, out_file=None):
    """
    Writes an image containing only the volumes with b-value specified at
    input
    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, ext = op.splitext(op.basename(in_dwi))
        if ext == ".gz":
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_file = op.abspath("%s_tsoi%s" % (fname, ext))

    im = nb.load(in_dwi, mmap=NUMPY_MMAP)
    dwidata = im.get_data()
    bvals = np.loadtxt(in_bval)

    if b == 'diff':
        selection = np.where(bvals != 0)
    elif b == 'nodiff':
        selection = np.where(bvals == 0)
    else:
        selection = np.where(bvals == b)

    extdata = np.squeeze(dwidata.take(selection, axis=3))
    hdr = im.header.copy()
    hdr.set_data_shape(extdata.shape)
    nb.Nifti1Image(extdata, im.affine, hdr).to_filename(out_file)
    return out_file


def hmc_split(in_file, in_bval, ref_num=0, lowbval=5.0):
    """
    Selects the reference and moving volumes from a dwi dataset
    for the purpose of HMC.
    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    from nipype.interfaces.base import isdefined
    from nipype.utils import NUMPY_MMAP

    im = nb.load(in_file, mmap=NUMPY_MMAP)
    data = im.get_data()
    hdr = im.header.copy()
    bval = np.loadtxt(in_bval)

    lowbs = np.where(bval <= lowbval)[0]

    volid = lowbs[0]
    if (isdefined(ref_num) and (ref_num < len(lowbs))):
        volid = ref_num

    if volid == 0:
        data = data[..., 1:]
        bval = bval[1:]
    elif volid == (data.shape[-1] - 1):
        data = data[..., :-1]
        bval = bval[:-1]
    else:
        data = np.concatenate(
            (data[..., :volid], data[..., (volid + 1):]), axis=3)
        bval = np.hstack((bval[:volid], bval[(volid + 1):]))

    out_ref = op.abspath('hmc_ref.nii.gz')
    out_mov = op.abspath('hmc_mov.nii.gz')
    out_bval = op.abspath('bval_split.txt')

    refdata = data[..., volid]
    hdr.set_data_shape(refdata.shape)
    nb.Nifti1Image(refdata, im.affine, hdr).to_filename(out_ref)

    hdr.set_data_shape(data.shape)
    nb.Nifti1Image(data, im.affine, hdr).to_filename(out_mov)
    np.savetxt(out_bval, bval)
    return [out_ref, out_mov, out_bval, volid]


def remove_comp(in_file, in_bval, volid=0, out_file=None):
    """
    Removes the volume ``volid`` from the 4D nifti file
    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, ext = op.splitext(op.basename(in_file))
        if ext == ".gz":
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_file = op.abspath("%s_extract%s" % (fname, ext))

    im = nb.load(in_file, mmap=NUMPY_MMAP)
    data = im.get_data()
    hdr = im.header.copy()
    bval = np.loadtxt(in_bval)

    if volid == 0:
        data = data[..., 1:]
        bval = bval[1:]
    elif volid == (data.shape[-1] - 1):
        data = data[..., :-1]
        bval = bval[:-1]
    else:
        data = np.concatenate(
            (data[..., :volid], data[..., (volid + 1):]), axis=3)
        bval = np.hstack((bval[:volid], bval[(volid + 1):]))
    hdr.set_data_shape(data.shape)
    nb.Nifti1Image(data, im.affine, hdr).to_filename(out_file)

    out_bval = op.abspath('bval_extract.txt')
    np.savetxt(out_bval, bval)
    return out_file, out_bval


def insert_mat(inlist, volid=0):
    import numpy as np
    import os.path as op
    idfname = op.abspath('identity.mat')
    out = inlist
    np.savetxt(idfname, np.eye(4))
    out.insert(volid, idfname)
    return out


def recompose_dwi(in_dwi, in_bval, in_corrected, out_file=None):
    """
    Recompose back the dMRI data accordingly the b-values table after EC
    correction
    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, ext = op.splitext(op.basename(in_dwi))
        if ext == ".gz":
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_file = op.abspath("%s_eccorrect%s" % (fname, ext))

    im = nb.load(in_dwi, mmap=NUMPY_MMAP)
    dwidata = im.get_data()
    bvals = np.loadtxt(in_bval)
    dwis = np.where(bvals != 0)[0].tolist()

    if len(dwis) != len(in_corrected):
        raise RuntimeError(('Length of DWIs in b-values table and after'
                            'correction should match'))

    for bindex, dwi in zip(dwis, in_corrected):
        dwidata[..., bindex] = nb.load(dwi, mmap=NUMPY_MMAP).get_data()

    nb.Nifti1Image(dwidata, im.affine, im.header).to_filename(out_file)
    return out_file


def recompose_xfm(in_bval, in_xfms):
    """
    Insert identity transformation matrices in b0 volumes to build up a list
    """
    import numpy as np
    import os.path as op

    bvals = np.loadtxt(in_bval)
    xfms = iter([np.loadtxt(xfm) for xfm in in_xfms])
    out_files = []

    for i, b in enumerate(bvals):
        if b == 0.0:
            mat = np.eye(4)
        else:
            mat = next(xfms)

        out_name = op.abspath('eccor_%04d.mat' % i)
        out_files.append(out_name)
        np.savetxt(out_name, mat)

    return out_files


def time_avg(in_file, index=[0], out_file=None):
    """
    Average the input time-series, selecting the indices given in index

    .. warning:: time steps should be already registered (corrected for
      head motion artifacts).

    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, ext = op.splitext(op.basename(in_file))
        if ext == ".gz":
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_file = op.abspath("%s_baseline%s" % (fname, ext))

    index = np.atleast_1d(index).tolist()

    imgs = np.array(nb.four_to_three(nb.load(in_file, mmap=NUMPY_MMAP)))[index]
    if len(index) == 1:
        data = imgs[0].get_data().astype(np.float32)
    else:
        data = np.average(
            np.array([im.get_data().astype(np.float32) for im in imgs]),
            axis=0)

    hdr = imgs[0].header.copy()
    hdr.set_data_shape(data.shape)
    hdr.set_xyzt_units('mm')
    hdr.set_data_dtype(np.float32)
    nb.Nifti1Image(data, imgs[0].affine, hdr).to_filename(out_file)
    return out_file


def b0_indices(in_bval, max_b=10.0):
    """
    Extract the indices of slices in a b-values file with a low b value
    """
    import numpy as np
    bval = np.loadtxt(in_bval)
    return np.argwhere(bval <= max_b).flatten().tolist()


def b0_average(in_dwi, in_bval, max_b=10.0, out_file=None):
    """
    A function that averages the *b0* volumes from a DWI dataset.
    As current dMRI data are being acquired with all b-values > 0.0,
    the *lowb* volumes are selected by specifying the parameter max_b.

    .. warning:: *b0* should be already registered (head motion artifact should
      be corrected).

    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, ext = op.splitext(op.basename(in_dwi))
        if ext == ".gz":
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_file = op.abspath("%s_avg_b0%s" % (fname, ext))

    imgs = np.array(nb.four_to_three(nb.load(in_dwi, mmap=NUMPY_MMAP)))
    bval = np.loadtxt(in_bval)
    index = np.argwhere(bval <= max_b).flatten().tolist()

    b0s = [im.get_data().astype(np.float32) for im in imgs[index]]
    b0 = np.average(np.array(b0s), axis=0)

    hdr = imgs[0].header.copy()
    hdr.set_data_shape(b0.shape)
    hdr.set_xyzt_units('mm')
    hdr.set_data_dtype(np.float32)
    nb.Nifti1Image(b0, imgs[0].affine, hdr).to_filename(out_file)
    return out_file


def rotate_bvecs(in_bvec, in_matrix):
    """
    Rotates the input bvec file accordingly with a list of matrices.

    .. note:: the input affine matrix transforms points in the destination
      image to their corresponding coordinates in the original image.
      Therefore, this matrix should be inverted first, as we want to know
      the target position of :math:`\\vec{r}`.

    """
    import os
    import numpy as np

    name, fext = os.path.splitext(os.path.basename(in_bvec))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('%s_rotated.bvec' % name)
    bvecs = np.loadtxt(in_bvec).T
    new_bvecs = []

    if len(bvecs) != len(in_matrix):
        raise RuntimeError(('Number of b-vectors (%d) and rotation '
                            'matrices (%d) should match.') % (len(bvecs),
                                                              len(in_matrix)))

    for bvec, mat in zip(bvecs, in_matrix):
        if np.all(bvec == 0.0):
            new_bvecs.append(bvec)
        else:
            invrot = np.linalg.inv(np.loadtxt(mat))[:3, :3]
            newbvec = invrot.dot(bvec)
            new_bvecs.append((newbvec / np.linalg.norm(newbvec)))

    np.savetxt(out_file, np.array(new_bvecs).T, fmt=b'%0.15f')
    return out_file


def eddy_rotate_bvecs(in_bvec, eddy_params):
    """
    Rotates the input bvec file accordingly with a list of parameters sourced
    from ``eddy``, as explained `here
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/EDDY/Faq#Will_eddy_rotate_my_bevcs_for_me.3F>`_.
    """
    import os
    import numpy as np
    from math import sin, cos

    name, fext = os.path.splitext(os.path.basename(in_bvec))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('%s_rotated.bvec' % name)
    bvecs = np.loadtxt(in_bvec).T
    new_bvecs = []

    params = np.loadtxt(eddy_params)

    if len(bvecs) != len(params):
        raise RuntimeError(('Number of b-vectors and rotation '
                            'matrices should match.'))

    for bvec, row in zip(bvecs, params):
        if np.all(bvec == 0.0):
            new_bvecs.append(bvec)
        else:
            ax = row[3]
            ay = row[4]
            az = row[5]

            Rx = np.array([[1.0, 0.0, 0.0], [0.0, cos(ax), -sin(ax)],
                           [0.0, sin(ax), cos(ax)]])
            Ry = np.array([[cos(ay), 0.0, sin(ay)], [0.0, 1.0, 0.0],
                           [-sin(ay), 0.0, cos(ay)]])
            Rz = np.array([[cos(az), -sin(az), 0.0], [sin(az),
                                                      cos(az), 0.0],
                           [0.0, 0.0, 1.0]])
            R = Rx.dot(Ry).dot(Rz)

            invrot = np.linalg.inv(R)
            newbvec = invrot.dot(bvec)
            new_bvecs.append(newbvec / np.linalg.norm(newbvec))

    np.savetxt(out_file, np.array(new_bvecs).T, fmt=b'%0.15f')
    return out_file


def compute_readout(params):
    """
    Computes readout time from epi params (see `eddy documentation
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/EDDY/Faq#How_do_I_know_what_to_put_into_my_--acqp_file.3F>`_).

    .. warning:: ``params['echospacing']`` should be in *sec* units.


    """
    epi_factor = 1.0
    acc_factor = 1.0
    try:
        if params['epi_factor'] > 1:
            epi_factor = float(params['epi_factor'] - 1)
    except:
        pass
    try:
        if params['acc_factor'] > 1:
            acc_factor = 1.0 / params['acc_factor']
    except:
        pass
    return acc_factor * epi_factor * params['echospacing']


def siemens2rads(in_file, out_file=None):
    """
    Converts input phase difference map to rads
    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    import math

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == '.gz':
            fname, _ = op.splitext(fname)
        out_file = op.abspath('./%s_rads.nii.gz' % fname)

    in_file = np.atleast_1d(in_file).tolist()
    im = nb.load(in_file[0])
    data = im.get_data().astype(np.float32)
    hdr = im.header.copy()

    if len(in_file) == 2:
        data = nb.load(in_file[1]).get_data().astype(np.float32) - data
    elif (data.ndim == 4) and (data.shape[-1] == 2):
        data = np.squeeze(data[..., 1] - data[..., 0])
        hdr.set_data_shape(data.shape[:3])

    imin = data.min()
    imax = data.max()
    data = (2.0 * math.pi * (data - imin) / (imax - imin)) - math.pi
    hdr.set_data_dtype(np.float32)
    hdr.set_xyzt_units('mm')
    hdr['datatype'] = 16
    nb.Nifti1Image(data, im.affine, hdr).to_filename(out_file)
    return out_file


def rads2radsec(in_file, delta_te, out_file=None):
    """
    Converts input phase difference map to rads
    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    import math
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == '.gz':
            fname, _ = op.splitext(fname)
        out_file = op.abspath('./%s_radsec.nii.gz' % fname)

    im = nb.load(in_file, mmap=NUMPY_MMAP)
    data = im.get_data().astype(np.float32) * (1.0 / delta_te)
    nb.Nifti1Image(data, im.affine, im.header).to_filename(out_file)
    return out_file


def demean_image(in_file, in_mask=None, out_file=None):
    """
    Demean image data inside mask
    """
    import numpy as np
    import nibabel as nb
    import os.path as op
    import math
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == '.gz':
            fname, _ = op.splitext(fname)
        out_file = op.abspath('./%s_demean.nii.gz' % fname)

    im = nb.load(in_file, mmap=NUMPY_MMAP)
    data = im.get_data().astype(np.float32)
    msk = np.ones_like(data)

    if in_mask is not None:
        msk = nb.load(in_mask, mmap=NUMPY_MMAP).get_data().astype(np.float32)
        msk[msk > 0] = 1.0
        msk[msk < 1] = 0.0

    mean = np.median(data[msk == 1].reshape(-1))
    data[msk == 1] = data[msk == 1] - mean
    nb.Nifti1Image(data, im.affine, im.header).to_filename(out_file)
    return out_file


def add_empty_vol(in_file, out_file=None):
    """
    Adds an empty vol to the phase difference image
    """
    import nibabel as nb
    import os.path as op
    import numpy as np
    import math
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == '.gz':
            fname, _ = op.splitext(fname)
        out_file = op.abspath('./%s_4D.nii.gz' % fname)

    im = nb.load(in_file, mmap=NUMPY_MMAP)
    zim = nb.Nifti1Image(np.zeros_like(im.get_data()), im.affine, im.header)
    nb.funcs.concat_images([im, zim]).to_filename(out_file)
    return out_file


def reorient_bvecs(in_dwi, old_dwi, in_bvec):
    """
    Checks reorientations of ``in_dwi`` w.r.t. ``old_dwi`` and
    reorients the in_bvec table accordingly.
    """
    import os
    import numpy as np
    import nibabel as nb
    from nipype.utils import NUMPY_MMAP

    name, fext = os.path.splitext(os.path.basename(in_bvec))
    if fext == '.gz':
        name, _ = os.path.splitext(name)
    out_file = os.path.abspath('%s_reorient.bvec' % name)
    bvecs = np.loadtxt(in_bvec).T
    new_bvecs = []

    N = nb.load(in_dwi, mmap=NUMPY_MMAP).affine
    O = nb.load(old_dwi, mmap=NUMPY_MMAP).affine
    RS = N.dot(np.linalg.inv(O))[:3, :3]
    sc_idx = np.where((np.abs(RS) != 1) & (RS != 0))
    S = np.ones_like(RS)
    S[sc_idx] = RS[sc_idx]
    R = RS / S

    new_bvecs = [R.dot(b) for b in bvecs]
    np.savetxt(out_file, np.array(new_bvecs).T, fmt=b'%0.15f')
    return out_file


def copy_hdr(in_file, in_file_hdr, out_file=None):
    import numpy as np
    import nibabel as nb
    import os.path as op
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == '.gz':
            fname, _ = op.splitext(fname)
        out_file = op.abspath('./%s_fixhdr.nii.gz' % fname)

    imref = nb.load(in_file_hdr, mmap=NUMPY_MMAP)
    hdr = imref.header.copy()
    hdr.set_data_dtype(np.float32)
    vsm = nb.load(in_file, mmap=NUMPY_MMAP).get_data().astype(np.float32)
    hdr.set_data_shape(vsm.shape)
    hdr.set_xyzt_units('mm')
    nii = nb.Nifti1Image(vsm, imref.affine, hdr)
    nii.to_filename(out_file)
    return out_file


def enhance(in_file, clip_limit=0.010, in_mask=None, out_file=None):
    import numpy as np
    import nibabel as nb
    import os.path as op
    from skimage import exposure, img_as_int
    from nipype.utils import NUMPY_MMAP

    if out_file is None:
        fname, fext = op.splitext(op.basename(in_file))
        if fext == '.gz':
            fname, _ = op.splitext(fname)
        out_file = op.abspath('./%s_enh.nii.gz' % fname)

    im = nb.load(in_file, mmap=NUMPY_MMAP)
    imdata = im.get_data()
    imshape = im.shape

    if in_mask is not None:
        msk = nb.load(in_mask, mmap=NUMPY_MMAP).get_data()
        msk[msk > 0] = 1
        msk[msk < 1] = 0
        imdata = imdata * msk

    immin = imdata.min()
    imdata = (imdata - immin).astype(np.uint16)

    adapted = exposure.equalize_adapthist(
        imdata.reshape(imshape[0], -1), clip_limit=clip_limit)

    nb.Nifti1Image(adapted.reshape(imshape), im.affine,
                   im.header).to_filename(out_file)

    return out_file


def _checkinitxfm(in_bval, excl_nodiff, in_xfms=None):
    from nipype.interfaces.base import isdefined
    import numpy as np
    import os.path as op
    bvals = np.loadtxt(in_bval)

    gen_id = ((in_xfms is None) or (not isdefined(in_xfms))
              or (len(in_xfms) != len(bvals)))

    init_xfms = []
    if excl_nodiff:
        dws = np.where(bvals != 0)[0].tolist()
    else:
        dws = list(range(len(bvals)))

    if gen_id:
        for i in dws:
            xfm_file = op.abspath('init_%04d.mat' % i)
            np.savetxt(xfm_file, np.eye(4))
            init_xfms.append(xfm_file)
    else:
        init_xfms = [in_xfms[i] for i in dws]

    return init_xfms



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import ConvertXFM


def test_ConvertXFM_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        concat_xfm=dict(
            argstr='-concat',
            position=-3,
            requires=['in_file2'],
            xor=['invert_xfm', 'concat_xfm', 'fix_scale_skew'],
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fix_scale_skew=dict(
            argstr='-fixscaleskew',
            position=-3,
            requires=['in_file2'],
            xor=['invert_xfm', 'concat_xfm', 'fix_scale_skew'],
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr='%s',
            mandatory=True,
            position=-1,
        ),
        in_file2=dict(
            argstr='%s',
            position=-2,
        ),
        invert_xfm=dict(
            argstr='-inverse',
            position=-3,
            xor=['invert_xfm', 'concat_xfm', 'fix_scale_skew'],
        ),
        out_file=dict(
            argstr='-omat %s',
            genfile=True,
            hash_files=False,
            position=1,
        ),
        output_type=dict(),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = ConvertXFM.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_ConvertXFM_outputs():
    output_map = dict(out_file=dict(), )
    outputs = ConvertXFM.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Tests for workflow callbacks
"""

from builtins import object

import pytest
import sys
import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe


def func():
    return


def bad_func():
    raise Exception


class Status(object):
    def __init__(self):
        self.statuses = []

    def callback(self, node, status, result=None):
        self.statuses.append((node, status))


def test_callback_normal(tmpdir):
    tmpdir.chdir()

    so = Status()
    wf = pe.Workflow(name='test', base_dir=tmpdir.strpath)
    f_node = pe.Node(
        niu.Function(function=func, input_names=[], output_names=[]),
        name='f_node')
    wf.add_nodes([f_node])
    wf.config['execution'] = {'crashdump_dir': wf.base_dir}
    wf.run(plugin="Linear", plugin_args={'status_callback': so.callback})
    assert len(so.statuses) == 2
    for (n, s) in so.statuses:
        assert n.name == 'f_node'
    assert so.statuses[0][1] == 'start'
    assert so.statuses[1][1] == 'end'


def test_callback_exception(tmpdir):
    tmpdir.chdir()

    so = Status()
    wf = pe.Workflow(name='test', base_dir=tmpdir.strpath)
    f_node = pe.Node(
        niu.Function(function=bad_func, input_names=[], output_names=[]),
        name='f_node')
    wf.add_nodes([f_node])
    wf.config['execution'] = {'crashdump_dir': wf.base_dir}
    try:
        wf.run(plugin="Linear", plugin_args={'status_callback': so.callback})
    except:
        pass
    assert len(so.statuses) == 2
    for (n, s) in so.statuses:
        assert n.name == 'f_node'
    assert so.statuses[0][1] == 'start'
    assert so.statuses[1][1] == 'exception'


def test_callback_multiproc_normal(tmpdir):
    tmpdir.chdir()

    so = Status()
    wf = pe.Workflow(name='test', base_dir=tmpdir.strpath)
    f_node = pe.Node(
        niu.Function(function=func, input_names=[], output_names=[]),
        name='f_node')
    wf.add_nodes([f_node])
    wf.config['execution']['crashdump_dir'] = wf.base_dir
    wf.config['execution']['poll_sleep_duration'] = 2
    wf.run(plugin='MultiProc', plugin_args={'status_callback': so.callback})
    assert len(so.statuses) == 2
    for (n, s) in so.statuses:
        assert n.name == 'f_node'
    assert so.statuses[0][1] == 'start'
    assert so.statuses[1][1] == 'end'


def test_callback_multiproc_exception(tmpdir):
    tmpdir.chdir()

    so = Status()
    wf = pe.Workflow(name='test', base_dir=tmpdir.strpath)
    f_node = pe.Node(
        niu.Function(function=bad_func, input_names=[], output_names=[]),
        name='f_node')
    wf.add_nodes([f_node])
    wf.config['execution'] = {'crashdump_dir': wf.base_dir}

    try:
        wf.run(
            plugin='MultiProc', plugin_args={
                'status_callback': so.callback
            })
    except:
        pass
    assert len(so.statuses) == 2
    for (n, s) in so.statuses:
        assert n.name == 'f_node'
    assert so.statuses[0][1] == 'start'
    assert so.statuses[1][1] == 'exception'



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from builtins import map, range


def get_vox_dims(volume):
    import nibabel as nb
    from nipype.utils import NUMPY_MMAP
    if isinstance(volume, list):
        volume = volume[0]
    nii = nb.load(volume, mmap=NUMPY_MMAP)
    hdr = nii.header
    voxdims = hdr.get_zooms()
    return [float(voxdims[0]), float(voxdims[1]), float(voxdims[2])]


def get_data_dims(volume):
    import nibabel as nb
    from nipype.utils import NUMPY_MMAP
    if isinstance(volume, list):
        volume = volume[0]
    nii = nb.load(volume, mmap=NUMPY_MMAP)
    hdr = nii.header
    datadims = hdr.get_data_shape()
    return [int(datadims[0]), int(datadims[1]), int(datadims[2])]


def get_affine(volume):
    import nibabel as nb
    from nipype.utils import NUMPY_MMAP
    nii = nb.load(volume, mmap=NUMPY_MMAP)
    return nii.affine


def select_aparc(list_of_files):
    for in_file in list_of_files:
        if 'aparc+aseg.mgz' in in_file:
            idx = list_of_files.index(in_file)
    return list_of_files[idx]


def select_aparc_annot(list_of_files):
    for in_file in list_of_files:
        if '.aparc.annot' in in_file:
            idx = list_of_files.index(in_file)
    return list_of_files[idx]


def region_list_from_volume(in_file):
    import nibabel as nb
    import numpy as np
    from nipype.utils import NUMPY_MMAP
    segmentation = nb.load(in_file, mmap=NUMPY_MMAP)
    segmentationdata = segmentation.get_data()
    rois = np.unique(segmentationdata)
    region_list = list(rois)
    region_list.sort()
    region_list.remove(0)
    region_list = list(map(int, region_list))
    return region_list


def id_list_from_lookup_table(lookup_file, region_list):
    import numpy as np
    LUTlabelsRGBA = np.loadtxt(
        lookup_file,
        skiprows=4,
        usecols=[0, 1, 2, 3, 4, 5],
        comments='#',
        dtype={
            'names': ('index', 'label', 'R', 'G', 'B', 'A'),
            'formats': ('int', '|S30', 'int', 'int', 'int', 'int')
        })
    numLUTLabels = np.size(LUTlabelsRGBA)
    LUTlabelDict = {}
    for labels in range(0, numLUTLabels):
        LUTlabelDict[LUTlabelsRGBA[labels][0]] = [
            LUTlabelsRGBA[labels][1], LUTlabelsRGBA[labels][2],
            LUTlabelsRGBA[labels][3], LUTlabelsRGBA[labels][4],
            LUTlabelsRGBA[labels][5]
        ]
    id_list = []
    for region in region_list:
        label = LUTlabelDict[region][0]
        id_list.append(label)
    id_list = list(map(str, id_list))
    return id_list



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..dti import ComputeMeanDiffusivity


def test_ComputeMeanDiffusivity_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr='< %s',
            mandatory=True,
            position=1,
        ),
        inputdatatype=dict(argstr='-inputdatatype %s', ),
        inputmodel=dict(argstr='-inputmodel %s', ),
        out_file=dict(
            argstr='> %s',
            genfile=True,
            position=-1,
        ),
        outputdatatype=dict(argstr='-outputdatatype %s', ),
        scheme_file=dict(
            argstr='%s',
            position=2,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = ComputeMeanDiffusivity.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_ComputeMeanDiffusivity_outputs():
    output_map = dict(md=dict(), )
    outputs = ComputeMeanDiffusivity.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..gtract import gtractInvertBSplineTransform


def test_gtractInvertBSplineTransform_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        inputReferenceVolume=dict(argstr='--inputReferenceVolume %s', ),
        inputTransform=dict(argstr='--inputTransform %s', ),
        landmarkDensity=dict(
            argstr='--landmarkDensity %s',
            sep=',',
        ),
        numberOfThreads=dict(argstr='--numberOfThreads %d', ),
        outputTransform=dict(
            argstr='--outputTransform %s',
            hash_files=False,
        ),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = gtractInvertBSplineTransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_gtractInvertBSplineTransform_outputs():
    output_map = dict(outputTransform=dict(), )
    outputs = gtractInvertBSplineTransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os
from distutils.version import LooseVersion

from .info import (LONG_DESCRIPTION as __doc__, URL as __url__, STATUS as
                   __status__, __version__)
from .utils.config import NipypeConfig
from .utils.logger import Logging
from .refs import due
from .pkg_info import get_pkg_info as _get_pkg_info

try:
    import faulthandler
    faulthandler.enable()
except (ImportError, IOError) as e:
    pass

config = NipypeConfig()
logging = Logging(config)


class NipypeTester(object):
    def __call__(self, doctests=True, parallel=True):
        try:
            import pytest
        except:
            raise RuntimeError(
                'py.test not installed, run: pip install pytest')
        args = []
        if not doctests:
            args.extend(['-p', 'no:doctest'])
        if not parallel:
            args.append('-n0')
        args.append(os.path.dirname(__file__))
        pytest.main(args=args)


test = NipypeTester()


def get_info():
    """Returns package information"""
    return _get_pkg_info(os.path.dirname(__file__))


from .pipeline import Node, MapNode, JoinNode, Workflow
from .interfaces import (DataGrabber, DataSink, SelectFiles, IdentityInterface,
                         Rename, Function, Select, Merge)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import division

import os
from ....interfaces import fsl as fsl  # fsl
from ....interfaces import utility as util  # utility
from ....pipeline import engine as pe  # pypeline engine
from ....interfaces import freesurfer as fs  # freesurfer
from ....interfaces import spm as spm
from ...smri.freesurfer.utils import create_getmask_flow
from .... import LooseVersion


def getthreshop(thresh):
    return ['-thr %.10f -Tmin -bin' % (0.1 * val[1]) for val in thresh]


def pickrun(files, whichrun):
    """pick file from list of files"""

    filemap = {'first': 0, 'last': -1, 'middle': len(files) // 2}

    if isinstance(files, list):

        # whichrun is given as integer
        if isinstance(whichrun, int):
            return files[whichrun]
        # whichrun is given as string
        elif isinstance(whichrun, str):
            if whichrun not in filemap.keys():
                raise (KeyError, 'Sorry, whichrun must be either integer index'
                       'or string in form of "first", "last" or "middle')
            else:
                return files[filemap[whichrun]]
    else:
        # in case single file name is given
        return files


def pickfirst(files):
    if isinstance(files, list):
        return files[0]
    else:
        return files


def pickmiddle(files):
    from nibabel import load
    import numpy as np
    from nipype.utils import NUMPY_MMAP
    middlevol = []
    for f in files:
        middlevol.append(int(np.ceil(load(f, mmap=NUMPY_MMAP).shape[3] / 2)))
    return middlevol


def pickvol(filenames, fileidx, which):
    from nibabel import load
    import numpy as np
    from nipype.utils import NUMPY_MMAP
    if which.lower() == 'first':
        idx = 0
    elif which.lower() == 'middle':
        idx = int(
            np.ceil(load(filenames[fileidx], mmap=NUMPY_MMAP).shape[3] / 2))
    elif which.lower() == 'last':
        idx = load(filenames[fileidx]).shape[3] - 1
    else:
        raise Exception('unknown value for volume selection : %s' % which)
    return idx


def getbtthresh(medianvals):
    return [0.75 * val for val in medianvals]


def chooseindex(fwhm):
    if fwhm < 1:
        return [0]
    else:
        return [1]


def getmeanscale(medianvals):
    return ['-mul %.10f' % (10000. / val) for val in medianvals]


def getusans(x):
    return [[tuple([val[0], 0.75 * val[1]])] for val in x]


tolist = lambda x: [x]
highpass_operand = lambda x: '-bptf %.10f -1' % x


def create_parallelfeat_preproc(name='featpreproc', highpass=True):
    """Preprocess each run with FSL independently of the others

    Parameters
    ----------

    ::

      name : name of workflow (default: featpreproc)
      highpass : boolean (default: True)

    Inputs::

        inputspec.func : functional runs (filename or list of filenames)
        inputspec.fwhm : fwhm for smoothing with SUSAN
        inputspec.highpass : HWHM in TRs (if created with highpass=True)

    Outputs::

        outputspec.reference : volume to which runs are realigned
        outputspec.motion_parameters : motion correction parameters
        outputspec.realigned_files : motion corrected files
        outputspec.motion_plots : plots of motion correction parameters
        outputspec.mask : mask file used to mask the brain
        outputspec.smoothed_files : smoothed functional data
        outputspec.highpassed_files : highpassed functional data (if highpass=True)
        outputspec.mean : mean file

    Example
    -------

    >>> preproc = create_parallelfeat_preproc()
    >>> preproc.inputs.inputspec.func = ['f3.nii', 'f5.nii']
    >>> preproc.inputs.inputspec.fwhm = 5
    >>> preproc.inputs.inputspec.highpass = 128./(2*2.5)
    >>> preproc.base_dir = '/tmp'
    >>> preproc.run() # doctest: +SKIP

    >>> preproc = create_parallelfeat_preproc(highpass=False)
    >>> preproc.inputs.inputspec.func = 'f3.nii'
    >>> preproc.inputs.inputspec.fwhm = 5
    >>> preproc.base_dir = '/tmp'
    >>> preproc.run() # doctest: +SKIP
    """
    version = 0
    if fsl.Info.version() and \
            LooseVersion(fsl.Info.version()) > LooseVersion('5.0.6'):
        version = 507

    featpreproc = pe.Workflow(name=name)
    """
    Set up a node to define all inputs required for the preprocessing workflow

    """

    if highpass:
        inputnode = pe.Node(
            interface=util.IdentityInterface(
                fields=['func', 'fwhm', 'highpass']),
            name='inputspec')
        outputnode = pe.Node(
            interface=util.IdentityInterface(fields=[
                'reference', 'motion_parameters', 'realigned_files',
                'motion_plots', 'mask', 'smoothed_files', 'highpassed_files',
                'mean'
            ]),
            name='outputspec')
    else:
        inputnode = pe.Node(
            interface=util.IdentityInterface(fields=['func', 'fwhm']),
            name='inputspec')
        outputnode = pe.Node(
            interface=util.IdentityInterface(fields=[
                'reference', 'motion_parameters', 'realigned_files',
                'motion_plots', 'mask', 'smoothed_files', 'mean'
            ]),
            name='outputspec')
    """
    Set up a node to define outputs for the preprocessing workflow

    """
    """
    Convert functional images to float representation. Since there can
    be more than one functional run we use a MapNode to convert each
    run.
    """

    img2float = pe.MapNode(
        interface=fsl.ImageMaths(
            out_data_type='float', op_string='', suffix='_dtype'),
        iterfield=['in_file'],
        name='img2float')
    featpreproc.connect(inputnode, 'func', img2float, 'in_file')
    """
    Extract the first volume of the first run as the reference
    """

    extract_ref = pe.MapNode(
        interface=fsl.ExtractROI(t_size=1),
        iterfield=['in_file', 't_min'],
        name='extractref')

    featpreproc.connect(img2float, 'out_file', extract_ref, 'in_file')
    featpreproc.connect(img2float, ('out_file', pickmiddle), extract_ref,
                        't_min')
    featpreproc.connect(extract_ref, 'roi_file', outputnode, 'reference')
    """
    Realign the functional runs to the reference (1st volume of first run)
    """

    motion_correct = pe.MapNode(
        interface=fsl.MCFLIRT(save_mats=True, save_plots=True),
        name='realign',
        iterfield=['in_file', 'ref_file'])
    featpreproc.connect(img2float, 'out_file', motion_correct, 'in_file')
    featpreproc.connect(extract_ref, 'roi_file', motion_correct, 'ref_file')
    featpreproc.connect(motion_correct, 'par_file', outputnode,
                        'motion_parameters')
    featpreproc.connect(motion_correct, 'out_file', outputnode,
                        'realigned_files')
    """
    Plot the estimated motion parameters
    """

    plot_motion = pe.MapNode(
        interface=fsl.PlotMotionParams(in_source='fsl'),
        name='plot_motion',
        iterfield=['in_file'])
    plot_motion.iterables = ('plot_type', ['rotations', 'translations'])
    featpreproc.connect(motion_correct, 'par_file', plot_motion, 'in_file')
    featpreproc.connect(plot_motion, 'out_file', outputnode, 'motion_plots')
    """
    Extract the mean volume of the first functional run
    """

    meanfunc = pe.MapNode(
        interface=fsl.ImageMaths(op_string='-Tmean', suffix='_mean'),
        iterfield=['in_file'],
        name='meanfunc')
    featpreproc.connect(motion_correct, 'out_file', meanfunc, 'in_file')
    """
    Strip the skull from the mean functional to generate a mask
    """

    meanfuncmask = pe.MapNode(
        interface=fsl.BET(mask=True, no_output=True, frac=0.3),
        iterfield=['in_file'],
        name='meanfuncmask')
    featpreproc.connect(meanfunc, 'out_file', meanfuncmask, 'in_file')
    """
    Mask the functional runs with the extracted mask
    """

    maskfunc = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_bet', op_string='-mas'),
        iterfield=['in_file', 'in_file2'],
        name='maskfunc')
    featpreproc.connect(motion_correct, 'out_file', maskfunc, 'in_file')
    featpreproc.connect(meanfuncmask, 'mask_file', maskfunc, 'in_file2')
    """
    Determine the 2nd and 98th percentile intensities of each functional run
    """

    getthresh = pe.MapNode(
        interface=fsl.ImageStats(op_string='-p 2 -p 98'),
        iterfield=['in_file'],
        name='getthreshold')
    featpreproc.connect(maskfunc, 'out_file', getthresh, 'in_file')
    """
    Threshold the first run of the functional data at 10% of the 98th percentile
    """

    threshold = pe.MapNode(
        interface=fsl.ImageMaths(out_data_type='char', suffix='_thresh'),
        iterfield=['in_file', 'op_string'],
        name='threshold')
    featpreproc.connect(maskfunc, 'out_file', threshold, 'in_file')
    """
    Define a function to get 10% of the intensity
    """

    featpreproc.connect(getthresh, ('out_stat', getthreshop), threshold,
                        'op_string')
    """
    Determine the median value of the functional runs using the mask
    """

    medianval = pe.MapNode(
        interface=fsl.ImageStats(op_string='-k %s -p 50'),
        iterfield=['in_file', 'mask_file'],
        name='medianval')
    featpreproc.connect(motion_correct, 'out_file', medianval, 'in_file')
    featpreproc.connect(threshold, 'out_file', medianval, 'mask_file')
    """
    Dilate the mask
    """

    dilatemask = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_dil', op_string='-dilF'),
        iterfield=['in_file'],
        name='dilatemask')
    featpreproc.connect(threshold, 'out_file', dilatemask, 'in_file')
    featpreproc.connect(dilatemask, 'out_file', outputnode, 'mask')
    """
    Mask the motion corrected functional runs with the dilated mask
    """

    maskfunc2 = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_mask', op_string='-mas'),
        iterfield=['in_file', 'in_file2'],
        name='maskfunc2')
    featpreproc.connect(motion_correct, 'out_file', maskfunc2, 'in_file')
    featpreproc.connect(dilatemask, 'out_file', maskfunc2, 'in_file2')
    """
    Smooth each run using SUSAN with the brightness threshold set to 75%
    of the median value for each run and a mask consituting the mean
    functional
    """

    smooth = create_susan_smooth()

    featpreproc.connect(inputnode, 'fwhm', smooth, 'inputnode.fwhm')
    featpreproc.connect(maskfunc2, 'out_file', smooth, 'inputnode.in_files')
    featpreproc.connect(dilatemask, 'out_file', smooth, 'inputnode.mask_file')
    """
    Mask the smoothed data with the dilated mask
    """

    maskfunc3 = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_mask', op_string='-mas'),
        iterfield=['in_file', 'in_file2'],
        name='maskfunc3')
    featpreproc.connect(smooth, 'outputnode.smoothed_files', maskfunc3,
                        'in_file')

    featpreproc.connect(dilatemask, 'out_file', maskfunc3, 'in_file2')

    concatnode = pe.Node(interface=util.Merge(2), name='concat')
    featpreproc.connect(maskfunc2, ('out_file', tolist), concatnode, 'in1')
    featpreproc.connect(maskfunc3, ('out_file', tolist), concatnode, 'in2')
    """
    The following nodes select smooth or unsmoothed data depending on the
    fwhm. This is because SUSAN defaults to smoothing the data with about the
    voxel size of the input data if the fwhm parameter is less than 1/3 of the
    voxel size.
    """
    selectnode = pe.Node(interface=util.Select(), name='select')

    featpreproc.connect(concatnode, 'out', selectnode, 'inlist')

    featpreproc.connect(inputnode, ('fwhm', chooseindex), selectnode, 'index')
    featpreproc.connect(selectnode, 'out', outputnode, 'smoothed_files')
    """
    Scale the median value of the run is set to 10000
    """

    meanscale = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_gms'),
        iterfield=['in_file', 'op_string'],
        name='meanscale')
    featpreproc.connect(selectnode, 'out', meanscale, 'in_file')
    """
    Define a function to get the scaling factor for intensity normalization
    """

    featpreproc.connect(medianval, ('out_stat', getmeanscale), meanscale,
                        'op_string')
    """
    Perform temporal highpass filtering on the data
    """

    if highpass:
        highpass = pe.MapNode(
            interface=fsl.ImageMaths(suffix='_tempfilt'),
            iterfield=['in_file'],
            name='highpass')
        featpreproc.connect(inputnode, ('highpass', highpass_operand),
                            highpass, 'op_string')
        featpreproc.connect(meanscale, 'out_file', highpass, 'in_file')

        if version < 507:
            featpreproc.connect(highpass, 'out_file', outputnode,
                                'highpassed_files')
        else:
            """
            Add back the mean removed by the highpass filter operation as of FSL 5.0.7
            """
            meanfunc4 = pe.MapNode(
                interface=fsl.ImageMaths(op_string='-Tmean', suffix='_mean'),
                iterfield=['in_file'],
                name='meanfunc4')

            featpreproc.connect(meanscale, 'out_file', meanfunc4, 'in_file')
            addmean = pe.MapNode(
                interface=fsl.BinaryMaths(operation='add'),
                iterfield=['in_file', 'operand_file'],
                name='addmean')
            featpreproc.connect(highpass, 'out_file', addmean, 'in_file')
            featpreproc.connect(meanfunc4, 'out_file', addmean, 'operand_file')
            featpreproc.connect(addmean, 'out_file', outputnode,
                                'highpassed_files')
    """
    Generate a mean functional image from the first run
    """

    meanfunc3 = pe.MapNode(
        interface=fsl.ImageMaths(op_string='-Tmean', suffix='_mean'),
        iterfield=['in_file'],
        name='meanfunc3')

    featpreproc.connect(meanscale, 'out_file', meanfunc3, 'in_file')
    featpreproc.connect(meanfunc3, 'out_file', outputnode, 'mean')

    return featpreproc


def create_featreg_preproc(name='featpreproc',
                           highpass=True,
                           whichvol='middle',
                           whichrun=0):
    """Create a FEAT preprocessing workflow with registration to one volume of the first run

    Parameters
    ----------

    ::

        name : name of workflow (default: featpreproc)
        highpass : boolean (default: True)
        whichvol : which volume of the first run to register to ('first', 'middle', 'last', 'mean')
        whichrun : which run to draw reference volume from (integer index or 'first', 'middle', 'last')

    Inputs::

        inputspec.func : functional runs (filename or list of filenames)
        inputspec.fwhm : fwhm for smoothing with SUSAN
        inputspec.highpass : HWHM in TRs (if created with highpass=True)

    Outputs::

        outputspec.reference : volume to which runs are realigned
        outputspec.motion_parameters : motion correction parameters
        outputspec.realigned_files : motion corrected files
        outputspec.motion_plots : plots of motion correction parameters
        outputspec.mask : mask file used to mask the brain
        outputspec.smoothed_files : smoothed functional data
        outputspec.highpassed_files : highpassed functional data (if highpass=True)
        outputspec.mean : mean file

    Example
    -------

    >>> preproc = create_featreg_preproc()
    >>> preproc.inputs.inputspec.func = ['f3.nii', 'f5.nii']
    >>> preproc.inputs.inputspec.fwhm = 5
    >>> preproc.inputs.inputspec.highpass = 128./(2*2.5)
    >>> preproc.base_dir = '/tmp'
    >>> preproc.run() # doctest: +SKIP

    >>> preproc = create_featreg_preproc(highpass=False, whichvol='mean')
    >>> preproc.inputs.inputspec.func = 'f3.nii'
    >>> preproc.inputs.inputspec.fwhm = 5
    >>> preproc.base_dir = '/tmp'
    >>> preproc.run() # doctest: +SKIP
    """

    version = 0
    if fsl.Info.version() and \
            LooseVersion(fsl.Info.version()) > LooseVersion('5.0.6'):
        version = 507

    featpreproc = pe.Workflow(name=name)
    """
    Set up a node to define all inputs required for the preprocessing workflow

    """

    if highpass:
        inputnode = pe.Node(
            interface=util.IdentityInterface(
                fields=['func', 'fwhm', 'highpass']),
            name='inputspec')
        outputnode = pe.Node(
            interface=util.IdentityInterface(fields=[
                'reference', 'motion_parameters', 'realigned_files',
                'motion_plots', 'mask', 'smoothed_files', 'highpassed_files',
                'mean'
            ]),
            name='outputspec')
    else:
        inputnode = pe.Node(
            interface=util.IdentityInterface(fields=['func', 'fwhm']),
            name='inputspec')
        outputnode = pe.Node(
            interface=util.IdentityInterface(fields=[
                'reference', 'motion_parameters', 'realigned_files',
                'motion_plots', 'mask', 'smoothed_files', 'mean'
            ]),
            name='outputspec')
    """
    Set up a node to define outputs for the preprocessing workflow

    """
    """
    Convert functional images to float representation. Since there can
    be more than one functional run we use a MapNode to convert each
    run.
    """

    img2float = pe.MapNode(
        interface=fsl.ImageMaths(
            out_data_type='float', op_string='', suffix='_dtype'),
        iterfield=['in_file'],
        name='img2float')
    featpreproc.connect(inputnode, 'func', img2float, 'in_file')
    """
    Extract the middle (or what whichvol points to) volume of the first run as the reference
    """

    if whichvol != 'mean':
        extract_ref = pe.Node(
            interface=fsl.ExtractROI(t_size=1),
            iterfield=['in_file'],
            name='extractref')
        featpreproc.connect(img2float, ('out_file', pickrun, whichrun),
                            extract_ref, 'in_file')
        featpreproc.connect(img2float, ('out_file', pickvol, 0, whichvol),
                            extract_ref, 't_min')
        featpreproc.connect(extract_ref, 'roi_file', outputnode, 'reference')
    """
    Realign the functional runs to the reference (`whichvol` volume of first run)
    """

    motion_correct = pe.MapNode(
        interface=fsl.MCFLIRT(
            save_mats=True, save_plots=True, interpolation='spline'),
        name='realign',
        iterfield=['in_file'])
    featpreproc.connect(img2float, 'out_file', motion_correct, 'in_file')
    if whichvol != 'mean':
        featpreproc.connect(extract_ref, 'roi_file', motion_correct,
                            'ref_file')
    else:
        motion_correct.inputs.mean_vol = True
        featpreproc.connect(motion_correct, ('mean_img', pickrun, whichrun),
                            outputnode, 'reference')

    featpreproc.connect(motion_correct, 'par_file', outputnode,
                        'motion_parameters')
    featpreproc.connect(motion_correct, 'out_file', outputnode,
                        'realigned_files')
    """
    Plot the estimated motion parameters
    """

    plot_motion = pe.MapNode(
        interface=fsl.PlotMotionParams(in_source='fsl'),
        name='plot_motion',
        iterfield=['in_file'])
    plot_motion.iterables = ('plot_type', ['rotations', 'translations'])
    featpreproc.connect(motion_correct, 'par_file', plot_motion, 'in_file')
    featpreproc.connect(plot_motion, 'out_file', outputnode, 'motion_plots')
    """
    Extract the mean volume of the first functional run
    """

    meanfunc = pe.Node(
        interface=fsl.ImageMaths(op_string='-Tmean', suffix='_mean'),
        name='meanfunc')
    featpreproc.connect(motion_correct, ('out_file', pickrun, whichrun),
                        meanfunc, 'in_file')
    """
    Strip the skull from the mean functional to generate a mask
    """

    meanfuncmask = pe.Node(
        interface=fsl.BET(mask=True, no_output=True, frac=0.3),
        name='meanfuncmask')
    featpreproc.connect(meanfunc, 'out_file', meanfuncmask, 'in_file')
    """
    Mask the functional runs with the extracted mask
    """

    maskfunc = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_bet', op_string='-mas'),
        iterfield=['in_file'],
        name='maskfunc')
    featpreproc.connect(motion_correct, 'out_file', maskfunc, 'in_file')
    featpreproc.connect(meanfuncmask, 'mask_file', maskfunc, 'in_file2')
    """
    Determine the 2nd and 98th percentile intensities of each functional run
    """

    getthresh = pe.MapNode(
        interface=fsl.ImageStats(op_string='-p 2 -p 98'),
        iterfield=['in_file'],
        name='getthreshold')
    featpreproc.connect(maskfunc, 'out_file', getthresh, 'in_file')
    """
    Threshold the first run of the functional data at 10% of the 98th percentile
    """

    threshold = pe.MapNode(
        interface=fsl.ImageMaths(out_data_type='char', suffix='_thresh'),
        iterfield=['in_file', 'op_string'],
        name='threshold')
    featpreproc.connect(maskfunc, 'out_file', threshold, 'in_file')
    """
    Define a function to get 10% of the intensity
    """

    featpreproc.connect(getthresh, ('out_stat', getthreshop), threshold,
                        'op_string')
    """
    Determine the median value of the functional runs using the mask
    """

    medianval = pe.MapNode(
        interface=fsl.ImageStats(op_string='-k %s -p 50'),
        iterfield=['in_file', 'mask_file'],
        name='medianval')
    featpreproc.connect(motion_correct, 'out_file', medianval, 'in_file')
    featpreproc.connect(threshold, 'out_file', medianval, 'mask_file')
    """
    Dilate the mask
    """

    dilatemask = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_dil', op_string='-dilF'),
        iterfield=['in_file'],
        name='dilatemask')
    featpreproc.connect(threshold, 'out_file', dilatemask, 'in_file')
    featpreproc.connect(dilatemask, 'out_file', outputnode, 'mask')
    """
    Mask the motion corrected functional runs with the dilated mask
    """

    maskfunc2 = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_mask', op_string='-mas'),
        iterfield=['in_file', 'in_file2'],
        name='maskfunc2')
    featpreproc.connect(motion_correct, 'out_file', maskfunc2, 'in_file')
    featpreproc.connect(dilatemask, 'out_file', maskfunc2, 'in_file2')
    """
    Smooth each run using SUSAN with the brightness threshold set to 75%
    of the median value for each run and a mask constituting the mean
    functional
    """

    smooth = create_susan_smooth()

    featpreproc.connect(inputnode, 'fwhm', smooth, 'inputnode.fwhm')
    featpreproc.connect(maskfunc2, 'out_file', smooth, 'inputnode.in_files')
    featpreproc.connect(dilatemask, 'out_file', smooth, 'inputnode.mask_file')
    """
    Mask the smoothed data with the dilated mask
    """

    maskfunc3 = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_mask', op_string='-mas'),
        iterfield=['in_file', 'in_file2'],
        name='maskfunc3')
    featpreproc.connect(smooth, 'outputnode.smoothed_files', maskfunc3,
                        'in_file')

    featpreproc.connect(dilatemask, 'out_file', maskfunc3, 'in_file2')

    concatnode = pe.Node(interface=util.Merge(2), name='concat')
    featpreproc.connect(maskfunc2, ('out_file', tolist), concatnode, 'in1')
    featpreproc.connect(maskfunc3, ('out_file', tolist), concatnode, 'in2')
    """
    The following nodes select smooth or unsmoothed data depending on the
    fwhm. This is because SUSAN defaults to smoothing the data with about the
    voxel size of the input data if the fwhm parameter is less than 1/3 of the
    voxel size.
    """
    selectnode = pe.Node(interface=util.Select(), name='select')

    featpreproc.connect(concatnode, 'out', selectnode, 'inlist')

    featpreproc.connect(inputnode, ('fwhm', chooseindex), selectnode, 'index')
    featpreproc.connect(selectnode, 'out', outputnode, 'smoothed_files')
    """
    Scale the median value of the run is set to 10000
    """

    meanscale = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_gms'),
        iterfield=['in_file', 'op_string'],
        name='meanscale')
    featpreproc.connect(selectnode, 'out', meanscale, 'in_file')
    """
    Define a function to get the scaling factor for intensity normalization
    """

    featpreproc.connect(medianval, ('out_stat', getmeanscale), meanscale,
                        'op_string')
    """
    Generate a mean functional image from the first run
    """

    meanfunc3 = pe.Node(
        interface=fsl.ImageMaths(op_string='-Tmean', suffix='_mean'),
        iterfield=['in_file'],
        name='meanfunc3')

    featpreproc.connect(meanscale, ('out_file', pickrun, whichrun), meanfunc3,
                        'in_file')
    featpreproc.connect(meanfunc3, 'out_file', outputnode, 'mean')
    """
    Perform temporal highpass filtering on the data
    """

    if highpass:
        highpass = pe.MapNode(
            interface=fsl.ImageMaths(suffix='_tempfilt'),
            iterfield=['in_file'],
            name='highpass')
        featpreproc.connect(inputnode, ('highpass', highpass_operand),
                            highpass, 'op_string')
        featpreproc.connect(meanscale, 'out_file', highpass, 'in_file')

        if version < 507:
            featpreproc.connect(highpass, 'out_file', outputnode,
                                'highpassed_files')
        else:
            """
            Add back the mean removed by the highpass filter operation as of FSL 5.0.7
            """
            meanfunc4 = pe.MapNode(
                interface=fsl.ImageMaths(op_string='-Tmean', suffix='_mean'),
                iterfield=['in_file'],
                name='meanfunc4')

            featpreproc.connect(meanscale, 'out_file', meanfunc4, 'in_file')
            addmean = pe.MapNode(
                interface=fsl.BinaryMaths(operation='add'),
                iterfield=['in_file', 'operand_file'],
                name='addmean')
            featpreproc.connect(highpass, 'out_file', addmean, 'in_file')
            featpreproc.connect(meanfunc4, 'out_file', addmean, 'operand_file')
            featpreproc.connect(addmean, 'out_file', outputnode,
                                'highpassed_files')

    return featpreproc


def create_susan_smooth(name="susan_smooth", separate_masks=True):
    """Create a SUSAN smoothing workflow

    Parameters
    ----------

    ::

        name : name of workflow (default: susan_smooth)
        separate_masks : separate masks for each run

    Inputs::

        inputnode.in_files : functional runs (filename or list of filenames)
        inputnode.fwhm : fwhm for smoothing with SUSAN (float or list of floats)
        inputnode.mask_file : mask used for estimating SUSAN thresholds (but not for smoothing)

    Outputs::

        outputnode.smoothed_files : functional runs (filename or list of filenames)

    Example
    -------

    >>> smooth = create_susan_smooth()
    >>> smooth.inputs.inputnode.in_files = 'f3.nii'
    >>> smooth.inputs.inputnode.fwhm = 5
    >>> smooth.inputs.inputnode.mask_file = 'mask.nii'
    >>> smooth.run() # doctest: +SKIP

    """

    # replaces the functionality of a "for loop"
    def cartesian_product(fwhms, in_files, usans, btthresh):
        from nipype.utils.filemanip import ensure_list
        # ensure all inputs are lists
        in_files = ensure_list(in_files)
        fwhms = [fwhms] if isinstance(fwhms, (int, float)) else fwhms
        # create cartesian product lists (s_<name> = single element of list)
        cart_in_file = [
            s_in_file for s_in_file in in_files for s_fwhm in fwhms
        ]
        cart_fwhm = [s_fwhm for s_in_file in in_files for s_fwhm in fwhms]
        cart_usans = [s_usans for s_usans in usans for s_fwhm in fwhms]
        cart_btthresh = [
            s_btthresh for s_btthresh in btthresh for s_fwhm in fwhms
        ]

        return cart_in_file, cart_fwhm, cart_usans, cart_btthresh

    susan_smooth = pe.Workflow(name=name)
    """
    Set up a node to define all inputs required for the preprocessing workflow

    """

    inputnode = pe.Node(
        interface=util.IdentityInterface(
            fields=['in_files', 'fwhm', 'mask_file']),
        name='inputnode')
    """
    Smooth each run using SUSAN with the brightness threshold set to 75%
    of the median value for each run and a mask consituting the mean
    functional
    """

    multi_inputs = pe.Node(
        util.Function(
            function=cartesian_product,
            output_names=[
                'cart_in_file', 'cart_fwhm', 'cart_usans', 'cart_btthresh'
            ]),
        name='multi_inputs')

    smooth = pe.MapNode(
        interface=fsl.SUSAN(),
        iterfield=['in_file', 'brightness_threshold', 'usans', 'fwhm'],
        name='smooth')
    """
    Determine the median value of the functional runs using the mask
    """

    if separate_masks:
        median = pe.MapNode(
            interface=fsl.ImageStats(op_string='-k %s -p 50'),
            iterfield=['in_file', 'mask_file'],
            name='median')
    else:
        median = pe.MapNode(
            interface=fsl.ImageStats(op_string='-k %s -p 50'),
            iterfield=['in_file'],
            name='median')
    susan_smooth.connect(inputnode, 'in_files', median, 'in_file')
    susan_smooth.connect(inputnode, 'mask_file', median, 'mask_file')
    """
    Mask the motion corrected functional runs with the dilated mask
    """

    if separate_masks:
        mask = pe.MapNode(
            interface=fsl.ImageMaths(suffix='_mask', op_string='-mas'),
            iterfield=['in_file', 'in_file2'],
            name='mask')
    else:
        mask = pe.MapNode(
            interface=fsl.ImageMaths(suffix='_mask', op_string='-mas'),
            iterfield=['in_file'],
            name='mask')
    susan_smooth.connect(inputnode, 'in_files', mask, 'in_file')
    susan_smooth.connect(inputnode, 'mask_file', mask, 'in_file2')
    """
    Determine the mean image from each functional run
    """

    meanfunc = pe.MapNode(
        interface=fsl.ImageMaths(op_string='-Tmean', suffix='_mean'),
        iterfield=['in_file'],
        name='meanfunc2')
    susan_smooth.connect(mask, 'out_file', meanfunc, 'in_file')
    """
    Merge the median values with the mean functional images into a coupled list
    """

    merge = pe.Node(interface=util.Merge(2, axis='hstack'), name='merge')
    susan_smooth.connect(meanfunc, 'out_file', merge, 'in1')
    susan_smooth.connect(median, 'out_stat', merge, 'in2')
    """
    Define a function to get the brightness threshold for SUSAN
    """

    susan_smooth.connect([
        (inputnode, multi_inputs, [('in_files', 'in_files'), ('fwhm',
                                                              'fwhms')]),
        (median, multi_inputs, [(('out_stat', getbtthresh), 'btthresh')]),
        (merge, multi_inputs, [(('out', getusans), 'usans')]),
        (multi_inputs, smooth,
         [('cart_in_file', 'in_file'), ('cart_fwhm', 'fwhm'),
          ('cart_btthresh', 'brightness_threshold'), ('cart_usans', 'usans')]),
    ])

    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=['smoothed_files']),
        name='outputnode')

    susan_smooth.connect(smooth, 'smoothed_file', outputnode, 'smoothed_files')

    return susan_smooth


def create_fsl_fs_preproc(name='preproc', highpass=True, whichvol='middle'):
    """Create a FEAT preprocessing workflow together with freesurfer

    Parameters
    ----------

    ::

        name : name of workflow (default: preproc)
        highpass : boolean (default: True)
        whichvol : which volume of the first run to register to ('first', 'middle', 'mean')

    Inputs::

        inputspec.func : functional runs (filename or list of filenames)
        inputspec.fwhm : fwhm for smoothing with SUSAN
        inputspec.highpass : HWHM in TRs (if created with highpass=True)
        inputspec.subject_id : freesurfer subject id
        inputspec.subjects_dir : freesurfer subjects dir

    Outputs::

        outputspec.reference : volume to which runs are realigned
        outputspec.motion_parameters : motion correction parameters
        outputspec.realigned_files : motion corrected files
        outputspec.motion_plots : plots of motion correction parameters
        outputspec.mask_file : mask file used to mask the brain
        outputspec.smoothed_files : smoothed functional data
        outputspec.highpassed_files : highpassed functional data (if highpass=True)
        outputspec.reg_file : bbregister registration files
        outputspec.reg_cost : bbregister registration cost files

    Example
    -------

    >>> preproc = create_fsl_fs_preproc(whichvol='first')
    >>> preproc.inputs.inputspec.highpass = 128./(2*2.5)
    >>> preproc.inputs.inputspec.func = ['f3.nii', 'f5.nii']
    >>> preproc.inputs.inputspec.subjects_dir = '.'
    >>> preproc.inputs.inputspec.subject_id = 's1'
    >>> preproc.inputs.inputspec.fwhm = 6
    >>> preproc.run() # doctest: +SKIP
    """

    featpreproc = pe.Workflow(name=name)
    """
    Set up a node to define all inputs required for the preprocessing workflow

    """

    if highpass:
        inputnode = pe.Node(
            interface=util.IdentityInterface(fields=[
                'func', 'fwhm', 'subject_id', 'subjects_dir', 'highpass'
            ]),
            name='inputspec')
        outputnode = pe.Node(
            interface=util.IdentityInterface(fields=[
                'reference', 'motion_parameters', 'realigned_files',
                'motion_plots', 'mask_file', 'smoothed_files',
                'highpassed_files', 'reg_file', 'reg_cost'
            ]),
            name='outputspec')
    else:
        inputnode = pe.Node(
            interface=util.IdentityInterface(
                fields=['func', 'fwhm', 'subject_id', 'subjects_dir']),
            name='inputspec')
        outputnode = pe.Node(
            interface=util.IdentityInterface(fields=[
                'reference', 'motion_parameters', 'realigned_files',
                'motion_plots', 'mask_file', 'smoothed_files', 'reg_file',
                'reg_cost'
            ]),
            name='outputspec')
    """
    Set up a node to define outputs for the preprocessing workflow

    """
    """
    Convert functional images to float representation. Since there can
    be more than one functional run we use a MapNode to convert each
    run.
    """

    img2float = pe.MapNode(
        interface=fsl.ImageMaths(
            out_data_type='float', op_string='', suffix='_dtype'),
        iterfield=['in_file'],
        name='img2float')
    featpreproc.connect(inputnode, 'func', img2float, 'in_file')
    """
    Extract the first volume of the first run as the reference
    """

    if whichvol != 'mean':
        extract_ref = pe.Node(
            interface=fsl.ExtractROI(t_size=1),
            iterfield=['in_file'],
            name='extractref')
        featpreproc.connect(img2float, ('out_file', pickfirst), extract_ref,
                            'in_file')
        featpreproc.connect(img2float, ('out_file', pickvol, 0, whichvol),
                            extract_ref, 't_min')
        featpreproc.connect(extract_ref, 'roi_file', outputnode, 'reference')
    """
    Realign the functional runs to the reference (1st volume of first run)
    """

    motion_correct = pe.MapNode(
        interface=fsl.MCFLIRT(
            save_mats=True, save_plots=True, interpolation='sinc'),
        name='realign',
        iterfield=['in_file'])
    featpreproc.connect(img2float, 'out_file', motion_correct, 'in_file')
    if whichvol != 'mean':
        featpreproc.connect(extract_ref, 'roi_file', motion_correct,
                            'ref_file')
    else:
        motion_correct.inputs.mean_vol = True
        featpreproc.connect(motion_correct, 'mean_img', outputnode,
                            'reference')

    featpreproc.connect(motion_correct, 'par_file', outputnode,
                        'motion_parameters')
    featpreproc.connect(motion_correct, 'out_file', outputnode,
                        'realigned_files')
    """
    Plot the estimated motion parameters
    """

    plot_motion = pe.MapNode(
        interface=fsl.PlotMotionParams(in_source='fsl'),
        name='plot_motion',
        iterfield=['in_file'])
    plot_motion.iterables = ('plot_type', ['rotations', 'translations'])
    featpreproc.connect(motion_correct, 'par_file', plot_motion, 'in_file')
    featpreproc.connect(plot_motion, 'out_file', outputnode, 'motion_plots')
    """Get the mask from subject for each run
    """

    maskflow = create_getmask_flow()
    featpreproc.connect([(inputnode, maskflow,
                          [('subject_id', 'inputspec.subject_id'),
                           ('subjects_dir', 'inputspec.subjects_dir')])])
    maskflow.inputs.inputspec.contrast_type = 't2'
    if whichvol != 'mean':
        featpreproc.connect(extract_ref, 'roi_file', maskflow,
                            'inputspec.source_file')
    else:
        featpreproc.connect(motion_correct, ('mean_img', pickfirst), maskflow,
                            'inputspec.source_file')
    """
    Mask the functional runs with the extracted mask
    """

    maskfunc = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_bet', op_string='-mas'),
        iterfield=['in_file'],
        name='maskfunc')
    featpreproc.connect(motion_correct, 'out_file', maskfunc, 'in_file')
    featpreproc.connect(maskflow, ('outputspec.mask_file', pickfirst),
                        maskfunc, 'in_file2')
    """
    Smooth each run using SUSAN with the brightness threshold set to 75%
    of the median value for each run and a mask consituting the mean
    functional
    """

    smooth = create_susan_smooth(separate_masks=False)

    featpreproc.connect(inputnode, 'fwhm', smooth, 'inputnode.fwhm')
    featpreproc.connect(maskfunc, 'out_file', smooth, 'inputnode.in_files')
    featpreproc.connect(maskflow, ('outputspec.mask_file', pickfirst), smooth,
                        'inputnode.mask_file')
    """
    Mask the smoothed data with the dilated mask
    """

    maskfunc3 = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_mask', op_string='-mas'),
        iterfield=['in_file'],
        name='maskfunc3')
    featpreproc.connect(smooth, 'outputnode.smoothed_files', maskfunc3,
                        'in_file')
    featpreproc.connect(maskflow, ('outputspec.mask_file', pickfirst),
                        maskfunc3, 'in_file2')

    concatnode = pe.Node(interface=util.Merge(2), name='concat')
    featpreproc.connect(maskfunc, ('out_file', tolist), concatnode, 'in1')
    featpreproc.connect(maskfunc3, ('out_file', tolist), concatnode, 'in2')
    """
    The following nodes select smooth or unsmoothed data depending on the
    fwhm. This is because SUSAN defaults to smoothing the data with about the
    voxel size of the input data if the fwhm parameter is less than 1/3 of the
    voxel size.
    """
    selectnode = pe.Node(interface=util.Select(), name='select')

    featpreproc.connect(concatnode, 'out', selectnode, 'inlist')

    featpreproc.connect(inputnode, ('fwhm', chooseindex), selectnode, 'index')
    featpreproc.connect(selectnode, 'out', outputnode, 'smoothed_files')
    """
    Scale the median value of the run is set to 10000
    """

    meanscale = pe.MapNode(
        interface=fsl.ImageMaths(suffix='_gms'),
        iterfield=['in_file', 'op_string'],
        name='meanscale')
    featpreproc.connect(selectnode, 'out', meanscale, 'in_file')
    """
    Determine the median value of the functional runs using the mask
    """

    medianval = pe.MapNode(
        interface=fsl.ImageStats(op_string='-k %s -p 50'),
        iterfield=['in_file'],
        name='medianval')
    featpreproc.connect(motion_correct, 'out_file', medianval, 'in_file')
    featpreproc.connect(maskflow, ('outputspec.mask_file', pickfirst),
                        medianval, 'mask_file')
    """
    Define a function to get the scaling factor for intensity normalization
    """

    featpreproc.connect(medianval, ('out_stat', getmeanscale), meanscale,
                        'op_string')
    """
    Perform temporal highpass filtering on the data
    """

    if highpass:
        highpass = pe.MapNode(
            interface=fsl.ImageMaths(suffix='_tempfilt'),
            iterfield=['in_file'],
            name='highpass')
        featpreproc.connect(inputnode, ('highpass', highpass_operand),
                            highpass, 'op_string')
        featpreproc.connect(meanscale, 'out_file', highpass, 'in_file')
        featpreproc.connect(highpass, 'out_file', outputnode,
                            'highpassed_files')

    featpreproc.connect(maskflow, ('outputspec.mask_file', pickfirst),
                        outputnode, 'mask_file')
    featpreproc.connect(maskflow, 'outputspec.reg_file', outputnode,
                        'reg_file')
    featpreproc.connect(maskflow, 'outputspec.reg_cost', outputnode,
                        'reg_cost')

    return featpreproc


def create_reg_workflow(name='registration'):
    """Create a FEAT preprocessing workflow

    Parameters
    ----------

    ::

        name : name of workflow (default: 'registration')

    Inputs::

        inputspec.source_files : files (filename or list of filenames to register)
        inputspec.mean_image : reference image to use
        inputspec.anatomical_image : anatomical image to coregister to
        inputspec.target_image : registration target

    Outputs::

        outputspec.func2anat_transform : FLIRT transform
        outputspec.anat2target_transform : FLIRT+FNIRT transform
        outputspec.transformed_files : transformed files in target space
        outputspec.transformed_mean : mean image in target space

    Example
    -------

    """

    register = pe.Workflow(name=name)

    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'source_files', 'mean_image', 'anatomical_image', 'target_image',
            'target_image_brain', 'config_file'
        ]),
        name='inputspec')
    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'func2anat_transform',
            'anat2target_transform',
            'transformed_files',
            'transformed_mean',
        ]),
        name='outputspec')
    """
    Estimate the tissue classes from the anatomical image. But use spm's segment
    as FSL appears to be breaking.
    """

    stripper = pe.Node(fsl.BET(), name='stripper')
    register.connect(inputnode, 'anatomical_image', stripper, 'in_file')
    fast = pe.Node(fsl.FAST(), name='fast')
    register.connect(stripper, 'out_file', fast, 'in_files')
    """
    Binarize the segmentation
    """

    binarize = pe.Node(
        fsl.ImageMaths(op_string='-nan -thr 0.5 -bin'), name='binarize')
    pickindex = lambda x, i: x[i]
    register.connect(fast, ('partial_volume_files', pickindex, 2), binarize,
                     'in_file')
    """
    Calculate rigid transform from mean image to anatomical image
    """

    mean2anat = pe.Node(fsl.FLIRT(), name='mean2anat')
    mean2anat.inputs.dof = 6
    register.connect(inputnode, 'mean_image', mean2anat, 'in_file')
    register.connect(stripper, 'out_file', mean2anat, 'reference')
    """
    Now use bbr cost function to improve the transform
    """

    mean2anatbbr = pe.Node(fsl.FLIRT(), name='mean2anatbbr')
    mean2anatbbr.inputs.dof = 6
    mean2anatbbr.inputs.cost = 'bbr'
    mean2anatbbr.inputs.schedule = os.path.join(
        os.getenv('FSLDIR'), 'etc/flirtsch/bbr.sch')
    register.connect(inputnode, 'mean_image', mean2anatbbr, 'in_file')
    register.connect(binarize, 'out_file', mean2anatbbr, 'wm_seg')
    register.connect(inputnode, 'anatomical_image', mean2anatbbr, 'reference')
    register.connect(mean2anat, 'out_matrix_file', mean2anatbbr,
                     'in_matrix_file')
    """
    Calculate affine transform from anatomical to target
    """

    anat2target_affine = pe.Node(fsl.FLIRT(), name='anat2target_linear')
    anat2target_affine.inputs.searchr_x = [-180, 180]
    anat2target_affine.inputs.searchr_y = [-180, 180]
    anat2target_affine.inputs.searchr_z = [-180, 180]
    register.connect(stripper, 'out_file', anat2target_affine, 'in_file')
    register.connect(inputnode, 'target_image_brain', anat2target_affine,
                     'reference')
    """
    Calculate nonlinear transform from anatomical to target
    """

    anat2target_nonlinear = pe.Node(fsl.FNIRT(), name='anat2target_nonlinear')
    anat2target_nonlinear.inputs.fieldcoeff_file = True
    register.connect(anat2target_affine, 'out_matrix_file',
                     anat2target_nonlinear, 'affine_file')
    register.connect(inputnode, 'anatomical_image', anat2target_nonlinear,
                     'in_file')
    register.connect(inputnode, 'config_file', anat2target_nonlinear,
                     'config_file')
    register.connect(inputnode, 'target_image', anat2target_nonlinear,
                     'ref_file')
    """
    Transform the mean image. First to anatomical and then to target
    """

    warpmean = pe.Node(fsl.ApplyWarp(interp='spline'), name='warpmean')
    register.connect(inputnode, 'mean_image', warpmean, 'in_file')
    register.connect(mean2anatbbr, 'out_matrix_file', warpmean, 'premat')
    register.connect(inputnode, 'target_image', warpmean, 'ref_file')
    register.connect(anat2target_nonlinear, 'fieldcoeff_file', warpmean,
                     'field_file')
    """
    Transform the remaining images. First to anatomical and then to target
    """

    warpall = pe.MapNode(
        fsl.ApplyWarp(interp='spline'),
        iterfield=['in_file'],
        nested=True,
        name='warpall')
    register.connect(inputnode, 'source_files', warpall, 'in_file')
    register.connect(mean2anatbbr, 'out_matrix_file', warpall, 'premat')
    register.connect(inputnode, 'target_image', warpall, 'ref_file')
    register.connect(anat2target_nonlinear, 'fieldcoeff_file', warpall,
                     'field_file')
    """
    Assign all the output files
    """

    register.connect(warpmean, 'out_file', outputnode, 'transformed_mean')
    register.connect(warpall, 'out_file', outputnode, 'transformed_files')
    register.connect(mean2anatbbr, 'out_matrix_file', outputnode,
                     'func2anat_transform')
    register.connect(anat2target_nonlinear, 'fieldcoeff_file', outputnode,
                     'anat2target_transform')

    return register



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Tests for the engine utils module
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import range, open

import os
import sys
from copy import deepcopy
import pytest

from ... import engine as pe
from ....interfaces import base as nib
from ....interfaces import utility as niu
from .... import config
from ..utils import clean_working_directory, write_workflow_prov


class InputSpec(nib.TraitedSpec):
    in_file = nib.File(exists=True, copyfile=True)


class OutputSpec(nib.TraitedSpec):
    output1 = nib.traits.List(nib.traits.Int, desc='outputs')


class UtilsTestInterface(nib.BaseInterface):
    input_spec = InputSpec
    output_spec = OutputSpec

    def _run_interface(self, runtime):
        runtime.returncode = 0
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['output1'] = [1]
        return outputs


def test_identitynode_removal(tmpdir):
    def test_function(arg1, arg2, arg3):
        import numpy as np
        return (np.array(arg1) + arg2 + arg3).tolist()

    wf = pe.Workflow(name="testidentity", base_dir=tmpdir.strpath)

    n1 = pe.Node(
        niu.IdentityInterface(fields=['a', 'b']),
        name='src',
        base_dir=tmpdir.strpath)
    n1.iterables = ('b', [0, 1, 2, 3])
    n1.inputs.a = [0, 1, 2, 3]

    n2 = pe.Node(niu.Select(), name='selector', base_dir=tmpdir.strpath)
    wf.connect(n1, ('a', test_function, 1, -1), n2, 'inlist')
    wf.connect(n1, 'b', n2, 'index')

    n3 = pe.Node(
        niu.IdentityInterface(fields=['c', 'd']),
        name='passer',
        base_dir=tmpdir.strpath)
    n3.inputs.c = [1, 2, 3, 4]
    wf.connect(n2, 'out', n3, 'd')

    n4 = pe.Node(niu.Select(), name='selector2', base_dir=tmpdir.strpath)
    wf.connect(n3, ('c', test_function, 1, -1), n4, 'inlist')
    wf.connect(n3, 'd', n4, 'index')

    fg = wf._create_flat_graph()
    wf._set_needed_outputs(fg)
    eg = pe.generate_expanded_graph(deepcopy(fg))
    assert len(eg.nodes()) == 8


def test_clean_working_directory(tmpdir):
    class OutputSpec(nib.TraitedSpec):
        files = nib.traits.List(nib.File)
        others = nib.File()

    class InputSpec(nib.TraitedSpec):
        infile = nib.File()

    outputs = OutputSpec()
    inputs = InputSpec()

    filenames = [
        'file.hdr', 'file.img', 'file.BRIK', 'file.HEAD', '_0x1234.json',
        'foo.txt'
    ]
    outfiles = []
    for filename in filenames:
        outfile = tmpdir.join(filename)
        outfile.write('dummy')
        outfiles.append(outfile.strpath)
    outputs.files = outfiles[:4:2]
    outputs.others = outfiles[5]
    inputs.infile = outfiles[-1]
    needed_outputs = ['files']
    config.set_default_config()
    assert os.path.exists(outfiles[5])
    config.set_default_config()
    config.set('execution', 'remove_unnecessary_outputs', False)
    out = clean_working_directory(outputs, tmpdir.strpath, inputs,
                                  needed_outputs, deepcopy(config._sections))
    assert os.path.exists(outfiles[5])
    assert out.others == outfiles[5]
    config.set('execution', 'remove_unnecessary_outputs', True)
    out = clean_working_directory(outputs, tmpdir.strpath, inputs,
                                  needed_outputs, deepcopy(config._sections))
    assert os.path.exists(outfiles[1])
    assert os.path.exists(outfiles[3])
    assert os.path.exists(outfiles[4])
    assert not os.path.exists(outfiles[5])
    assert out.others == nib.Undefined
    assert len(out.files) == 2
    config.set_default_config()


def create_wf(name):
    """Creates a workflow for the following tests"""
    def fwhm(fwhm):
        return fwhm

    pipe = pe.Workflow(name=name)
    process = pe.Node(
        niu.Function(
            input_names=['fwhm'], output_names=['fwhm'], function=fwhm),
        name='proc')
    process.iterables = ('fwhm', [0])
    process2 = pe.Node(
        niu.Function(
            input_names=['fwhm'], output_names=['fwhm'], function=fwhm),
        name='proc2')
    process2.iterables = ('fwhm', [0])
    pipe.connect(process, 'fwhm', process2, 'fwhm')
    return pipe


def test_multi_disconnected_iterable(tmpdir):
    metawf = pe.Workflow(name='meta')
    metawf.base_dir = tmpdir.strpath
    metawf.add_nodes([create_wf('wf%d' % i) for i in range(30)])
    eg = metawf.run(plugin='Linear')
    assert len(eg.nodes()) == 60


def test_provenance(tmpdir):
    metawf = pe.Workflow(name='meta')
    metawf.base_dir = tmpdir.strpath
    metawf.add_nodes([create_wf('wf%d' % i) for i in range(1)])
    eg = metawf.run(plugin='Linear')
    prov_base = tmpdir.join('workflow_provenance_test').strpath
    psg = write_workflow_prov(eg, prov_base, format='all')
    assert len(psg.bundles) == 2
    assert len(psg.get_records()) == 7


def dummy_func(value):
    return value + 1


@pytest.mark.skipif(
    sys.version_info < (3, 0), reason="the famous segfault #1788")
def test_mapnode_crash(tmpdir):
    """Test mapnode crash when stop_on_first_crash is True"""
    cwd = os.getcwd()
    node = pe.MapNode(
        niu.Function(
            input_names=['WRONG'],
            output_names=['newstring'],
            function=dummy_func),
        iterfield=['WRONG'],
        name='myfunc')
    node.inputs.WRONG = ['string{}'.format(i) for i in range(3)]
    node.config = deepcopy(config._sections)
    node.config['execution']['stop_on_first_crash'] = True
    node.base_dir = tmpdir.strpath
    with pytest.raises(TypeError):
        node.run()
    os.chdir(cwd)


@pytest.mark.skipif(
    sys.version_info < (3, 0), reason="the famous segfault #1788")
def test_mapnode_crash2(tmpdir):
    """Test mapnode crash when stop_on_first_crash is False"""
    cwd = os.getcwd()
    node = pe.MapNode(
        niu.Function(
            input_names=['WRONG'],
            output_names=['newstring'],
            function=dummy_func),
        iterfield=['WRONG'],
        name='myfunc')
    node.inputs.WRONG = ['string{}'.format(i) for i in range(3)]
    node.base_dir = tmpdir.strpath

    with pytest.raises(Exception):
        node.run()
    os.chdir(cwd)


@pytest.mark.skipif(
    sys.version_info < (3, 0), reason="the famous segfault #1788")
def test_mapnode_crash3(tmpdir):
    """Test mapnode crash when mapnode is embedded in a workflow"""
    tmpdir.chdir()
    node = pe.MapNode(
        niu.Function(
            input_names=['WRONG'],
            output_names=['newstring'],
            function=dummy_func),
        iterfield=['WRONG'],
        name='myfunc')
    node.inputs.WRONG = ['string{}'.format(i) for i in range(3)]
    wf = pe.Workflow('testmapnodecrash')
    wf.add_nodes([node])
    wf.base_dir = tmpdir.strpath
    # changing crashdump dir to cwl (to avoid problems with read-only systems)
    wf.config["execution"]["crashdump_dir"] = os.getcwd()
    with pytest.raises(RuntimeError):
        wf.run(plugin='Linear')



import pytest
import numpy
import os

DATADIR = os.path.realpath(
    os.path.join(os.path.dirname(__file__), 'testing/data'))


@pytest.fixture(autouse=True)
def add_np(doctest_namespace):
    doctest_namespace['np'] = numpy
    doctest_namespace['os'] = os

    doctest_namespace["datadir"] = DATADIR


@pytest.fixture(autouse=True)
def in_testing(request):
    # This seems to be a reliable way to distinguish tests from doctests
    if request.function is None:
        os.chdir(DATADIR)



# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from ....interfaces.utility import Function, IdentityInterface, Merge
from ....pipeline import engine as pe
from ....interfaces.freesurfer import *
from .utils import copy_file


def copy_ltas(in_file, subjects_dir, subject_id, long_template):
    import os
    out_file = copy_file(in_file,
                         os.path.basename(in_file).replace(
                             long_template, subject_id))
    return out_file


def create_AutoRecon2(name="AutoRecon2",
                      longitudinal=False,
                      plugin_args=None,
                      fsvernum=5.3,
                      stop=None,
                      shrink=None,
                      distance=None):
    # AutoRecon2
    # Workflow
    ar2_wf = pe.Workflow(name=name)

    inputspec = pe.Node(
        IdentityInterface(fields=[
            'orig',
            'nu',  # version < 6
            'brainmask',
            'transform',
            'subject_id',
            'template_talairach_lta',
            'template_talairach_m3z',
            'template_label_intensities',
            'template_aseg',
            'subj_to_template_lta',
            'alltps_to_template_ltas',
            'template_lh_white',
            'template_rh_white',
            'template_lh_pial',
            'template_rh_pial',
            'init_wm',
            'timepoints',
            'alltps_segs',
            'alltps_segs_noCC',
            'alltps_norms',
            'num_threads',
            'reg_template',
            'reg_template_withskull'
        ]),
        run_without_submitting=True,
        name='inputspec')

    # Input node
    if longitudinal:
        # TODO: Work on longitudinal workflow
        inputspec.inputs.timepoints = config['timepoints']

    if fsvernum >= 6:
        # NU Intensity Correction
        """
        Non-parametric Non-uniform intensity Normalization (N3), corrects for
        intensity non-uniformity in MR data, making relatively few assumptions about
        the data. This runs the MINC tool 'nu_correct'.
        """
        intensity_correction = pe.Node(
            MNIBiasCorrection(), name="Intensity_Correction")
        intensity_correction.inputs.out_file = 'nu.mgz'
        ar2_wf.connect([(inputspec, intensity_correction,
                         [('orig', 'in_file'), ('brainmask', 'mask'),
                          ('transform', 'transform')])])

        # intensity correction parameters are more specific in 6+
        intensity_correction.inputs.iterations = 1
        intensity_correction.inputs.protocol_iterations = 1000
        if stop:
            intensity_correction.inputs.stop = stop
        if shrink:
            intensity_correction.inputs.shrink = shrink
        intensity_correction.inputs.distance = distance

        add_to_header_nu = pe.Node(AddXFormToHeader(), name="Add_XForm_to_NU")
        add_to_header_nu.inputs.copy_name = True
        add_to_header_nu.inputs.out_file = 'nu.mgz'
        ar2_wf.connect([(intensity_correction, add_to_header_nu, [
            ('out_file', 'in_file'),
        ]), (inputspec, add_to_header_nu, [('transform', 'transform')])])

    # EM Registration
    """
    Computes the transform to align the mri/nu.mgz volume to the default GCA
    atlas found in FREESURFER_HOME/average (see -gca flag for more info).
    """
    if longitudinal:
        align_transform = pe.Node(
            Function(['in_file', 'out_file'], ['out_file'], copy_file),
            name='Copy_Talairach_lta')
        align_transform.inputs.out_file = 'talairach.lta'

        ar2_wf.connect([(inputspec, align_transform,
                         [('template_talairach_lta', 'in_file')])])
    else:
        align_transform = pe.Node(EMRegister(), name="Align_Transform")
        align_transform.inputs.out_file = 'talairach.lta'
        align_transform.inputs.nbrspacing = 3
        if plugin_args:
            align_transform.plugin_args = plugin_args
        ar2_wf.connect([(inputspec, align_transform,
                         [('brainmask', 'mask'), ('reg_template', 'template'),
                          ('num_threads', 'num_threads')])])
        if fsvernum >= 6:
            ar2_wf.connect([(add_to_header_nu, align_transform,
                             [('out_file', 'in_file')])])
        else:
            ar2_wf.connect([(inputspec, align_transform, [('nu', 'in_file')])])

    # CA Normalize
    """
    Further normalization, based on GCA model. The normalization is based on an
    estimate of the most certain segmentation voxels, which it then uses to
    estimate the bias field/scalings. Creates mri/norm.mgz.
    """
    ca_normalize = pe.Node(CANormalize(), name='CA_Normalize')
    ca_normalize.inputs.out_file = 'norm.mgz'
    if not longitudinal:
        ca_normalize.inputs.control_points = 'ctrl_pts.mgz'
    else:
        copy_template_aseg = pe.Node(
            Function(['in_file', 'out_file'], ['out_file'], copy_file),
            name='Copy_Template_Aseg')
        copy_template_aseg.inputs.out_file = 'aseg_{0}.mgz'.format(
            config['long_template'])

        ar1_wf.connect(
            [(inputspec, copy_template, [('template_aseg', 'in_file')]),
             (copy_template, ca_normalize, [('out_file', 'long_file')])])

    ar2_wf.connect([(align_transform, ca_normalize, [('out_file',
                                                      'transform')]),
                    (inputspec, ca_normalize, [('brainmask', 'mask'),
                                               ('reg_template', 'atlas')])])
    if fsvernum >= 6:
        ar2_wf.connect([(add_to_header_nu, ca_normalize, [('out_file',
                                                           'in_file')])])
    else:
        ar2_wf.connect([(inputspec, ca_normalize, [('nu', 'in_file')])])

    # CA Register
    # Computes a nonlinear transform to align with GCA atlas.
    ca_register = pe.Node(CARegister(), name='CA_Register')
    ca_register.inputs.align = 'after'
    ca_register.inputs.no_big_ventricles = True
    ca_register.inputs.out_file = 'talairach.m3z'
    if plugin_args:
        ca_register.plugin_args = plugin_args
    ar2_wf.connect([(ca_normalize, ca_register, [('out_file', 'in_file')]),
                    (inputspec, ca_register,
                     [('brainmask', 'mask'), ('num_threads', 'num_threads'),
                      ('reg_template', 'template')])])
    if not longitudinal:
        ar2_wf.connect([(align_transform, ca_register, [('out_file',
                                                         'transform')])])
    else:
        ca_register.inputs.levels = 2
        ca_register.inputs.A = 1
        ar2_wf.connect([(inputspec, ca_register, [('template_talairach_m3z',
                                                   'l_files')])])

    # Remove Neck
    """
    The neck region is removed from the NU-corrected volume mri/nu.mgz. Makes use
    of transform computed from prior CA Register stage.
    """
    remove_neck = pe.Node(RemoveNeck(), name='Remove_Neck')
    remove_neck.inputs.radius = 25
    remove_neck.inputs.out_file = 'nu_noneck.mgz'
    ar2_wf.connect([(ca_register, remove_neck, [('out_file', 'transform')]),
                    (inputspec, remove_neck, [('reg_template', 'template')])])
    if fsvernum >= 6:
        ar2_wf.connect([(add_to_header_nu, remove_neck, [('out_file',
                                                          'in_file')])])
    else:
        ar2_wf.connect([(inputspec, remove_neck, [('nu', 'in_file')])])

    # SkullLTA (EM Registration, with Skull)
    # Computes transform to align volume mri/nu_noneck.mgz with GCA volume
    # possessing the skull.
    em_reg_withskull = pe.Node(EMRegister(), name='EM_Register_withSkull')
    em_reg_withskull.inputs.skull = True
    em_reg_withskull.inputs.out_file = 'talairach_with_skull_2.lta'
    if plugin_args:
        em_reg_withskull.plugin_args = plugin_args
    ar2_wf.connect([(align_transform, em_reg_withskull, [('out_file',
                                                          'transform')]),
                    (remove_neck, em_reg_withskull, [('out_file', 'in_file')]),
                    (inputspec, em_reg_withskull,
                     [('num_threads', 'num_threads'),
                      ('reg_template_withskull', 'template')])])

    # SubCort Seg (CA Label)
    # Labels subcortical structures, based in GCA model.
    if longitudinal:
        copy_long_ltas = pe.MapNode(
            Function(
                ['in_file', 'subjects_dir', 'subject_id', 'long_template'],
                ['out_file'], copy_ltas),
            iterfield=['in_file'],
            name='Copy_long_ltas')
        ar2_wf.connect([(inputspec, copy_long_ltas,
                         [('alltps_to_template_ltas', 'in_file'),
                          ('subjects_dir', 'subjects_dir'), ('subject_id',
                                                             'subject_id')])])
        copy_long_ltas.inputs.long_template = config['long_template']

        merge_norms = pe.Node(Merge(2), name="Merge_Norms")

        ar2_wf.connect([(inputspec, merge_norms, [('alltps_norms', 'in1')]),
                        (ca_normalize, merge_norms, [('out_file', 'in2')])])

        fuse_segmentations = pe.Node(
            FuseSegmentations(), name="Fuse_Segmentations")

        ar2_wf.connect([(inputspec, fuse_segmentations, [
            ('timepoints', 'timepoints'), ('alltps_segs', 'in_segmentations'),
            ('alltps_segs_noCC', 'in_segmentations_noCC'), ('subject_id',
                                                            'subject_id')
        ]), (merge_norms, fuse_segmentations, [('out', 'in_norms')])])
        fuse_segmentations.inputs.out_file = 'aseg.fused.mgz'

    ca_label = pe.Node(CALabel(), name='CA_Label')
    if fsvernum >= 6:
        ca_label.inputs.relabel_unlikely = (9, .3)
        ca_label.inputs.prior = 0.5
    ca_label.inputs.align = True
    ca_label.inputs.out_file = 'aseg.auto_noCCseg.mgz'
    if plugin_args:
        ca_label.plugin_args = plugin_args
    ar2_wf.connect([(ca_normalize, ca_label, [('out_file', 'in_file')]),
                    (ca_register, ca_label, [('out_file', 'transform')]),
                    (inputspec, ca_label, [('num_threads', 'num_threads'),
                                           ('reg_template', 'template')])])

    if longitudinal:
        ar2_wf.connect([(fuse_segmentations, ca_label, [('out_file',
                                                         'in_vol')]),
                        (inputspec, ca_label, [('template_label_intensities',
                                                'intensities')])])

    # mri_cc - segments the corpus callosum into five separate labels in the
    # subcortical segmentation volume 'aseg.mgz'
    segment_cc = pe.Node(SegmentCC(), name="Segment_CorpusCallosum")
    segment_cc.inputs.out_rotation = 'cc_up.lta'
    segment_cc.inputs.out_file = 'aseg.auto.mgz'
    segment_cc.inputs.copy_inputs = True
    ar2_wf.connect([
        (ca_label, segment_cc, [('out_file', 'in_file')]),
        (ca_normalize, segment_cc, [('out_file', 'in_norm')]),
    ])

    copy_cc = pe.Node(
        Function(['in_file', 'out_file'], ['out_file'], copy_file),
        name='Copy_CCSegmentation')
    copy_cc.inputs.out_file = 'aseg.presurf.mgz'

    ar2_wf.connect([(segment_cc, copy_cc, [('out_file', 'in_file')])])

    # Normalization2
    """
    Performs a second (major) intensity correction using only the brain volume a
    s the input (so that it has to be done after the skull strip). Intensity
    normalization works better when the skull has been removed. Creates a new
    brain.mgz volume. The -autorecon2-cp stage begins here.
    """
    normalization2 = pe.Node(Normalize(), name="Normalization2")
    normalization2.inputs.out_file = 'brain.mgz'
    ar2_wf.connect([(copy_cc, normalization2, [('out_file', 'segmentation')]),
                    (inputspec, normalization2, [('brainmask', 'mask')]),
                    (ca_normalize, normalization2, [('out_file', 'in_file')])])

    # Mask Brain Final Surface

    # Applies brainmask.mgz to brain.mgz to create brain.finalsurfs.mgz.
    mri_mask = pe.Node(ApplyMask(), name="Mask_Brain_Final_Surface")
    mri_mask.inputs.mask_thresh = 5
    mri_mask.inputs.out_file = 'brain.finalsurfs.mgz'

    ar2_wf.connect([(normalization2, mri_mask, [('out_file', 'in_file')]),
                    (inputspec, mri_mask, [('brainmask', 'mask_file')])])

    # WM Segmentation
    """
    Attempts to separate white matter from everything else. The input is
    mri/brain.mgz, and the output is mri/wm.mgz. Uses intensity, neighborhood,
    and smoothness constraints. This is the volume that is edited when manually
    fixing defects. Calls mri_segment, mri_edit_wm_with_aseg, and mri_pretess.
    """

    wm_seg = pe.Node(SegmentWM(), name="Segment_WM")
    wm_seg.inputs.out_file = 'wm.seg.mgz'
    ar2_wf.connect([(normalization2, wm_seg, [('out_file', 'in_file')])])

    edit_wm = pe.Node(EditWMwithAseg(), name='Edit_WhiteMatter')
    edit_wm.inputs.out_file = 'wm.asegedit.mgz'
    edit_wm.inputs.keep_in = True
    ar2_wf.connect([(wm_seg, edit_wm, [('out_file', 'in_file')]),
                    (copy_cc, edit_wm, [('out_file', 'seg_file')]),
                    (normalization2, edit_wm, [('out_file', 'brain_file')])])

    pretess = pe.Node(MRIPretess(), name="MRI_Pretess")
    pretess.inputs.out_file = 'wm.mgz'
    pretess.inputs.label = 'wm'
    ar2_wf.connect([(edit_wm, pretess, [('out_file', 'in_filled')]),
                    (ca_normalize, pretess, [('out_file', 'in_norm')])])

    if longitudinal:
        transfer_init_wm = pe.Node(ApplyMask(), name="Transfer_Initial_WM")
        transfer_init_wm.inputs.transfer = 255
        transfer_init_wm.inputs.keep_mask_deletion_edits = True
        transfer_init_wm.inputs.out_file = 'wm.mgz'
        ar2_wf.connect([(pretess, transfer_init_wm, [('out_file', 'in_file')]),
                        (inputspec, transfer_init_wm,
                         [('init_wm', 'mask_file'), ('subj_to_template_lta',
                                                     'xfm_file')])])
        # changing the pretess variable so that the rest of the connections still work!!!
        pretess = transfer_init_wm

    # Fill
    """ This creates the subcortical mass from which the orig surface is created.
    The mid brain is cut from the cerebrum, and the hemispheres are cut from each
    other. The left hemisphere is binarized to 255. The right hemisphere is binarized
    to 127. The input is mri/wm.mgz and the output is mri/filled.mgz. Calls mri_fill.
    """

    fill = pe.Node(MRIFill(), name="Fill")
    fill.inputs.log_file = 'ponscc.cut.log'
    fill.inputs.out_file = 'filled.mgz'

    ar2_wf.connect([
        (pretess, fill, [('out_file', 'in_file')]),
        (align_transform, fill, [('out_file', 'transform')]),
        (ca_label, fill, [('out_file', 'segmentation')]),
    ])

    ar2_lh = pe.Workflow("AutoRecon2_Left")
    ar2_rh = pe.Workflow("AutoRecon2_Right")

    # iterate by hemisphere
    for hemisphere in ['lh', 'rh']:
        if hemisphere == 'lh':
            label = 255
            hemi_wf = ar2_lh
        else:
            label = 127
            hemi_wf = ar2_rh

        hemi_inputspec = pe.Node(
            IdentityInterface(fields=[
                'norm', 'filled', 'aseg', 't1', 'wm', 'brain', 'num_threads'
            ]),
            name="inputspec")

        if longitudinal:
            # Make White Surf
            # Copy files from longitudinal base
            copy_template_white = pe.Node(
                Function(['in_file', 'out_file'], ['out_file'], copy_file),
                name='Copy_Template_White')
            copy_template_white.inputs.out_file = '{0}.orig'.format(hemisphere)

            copy_template_orig_white = pe.Node(
                Function(['in_file', 'out_file'], ['out_file'], copy_file),
                name='Copy_Template_Orig_White')
            copy_template_orig_white.inputs.out_file = '{0}.orig_white'.format(
                hemisphere)

            copy_template_orig_pial = pe.Node(
                Function(['in_file', 'out_file'], ['out_file'], copy_file),
                name='Copy_Template_Orig_Pial')
            copy_template_orig_pial.inputs.out_file = '{0}.orig_pial'.format(
                hemisphere)

            # White

            # This function implicitly calls other inputs based on the subject_id
            # wf attempts to make sure files are data sinked to the correct
            # folders before calling
            make_surfaces = pe.Node(MakeSurfaces(), name="Make_Surfaces")
            make_surfaces.inputs.noaparc = True
            make_surfaces.inputs.mgz = True
            make_surfaces.inputs.white_only = True
            make_surfaces.inputs.hemisphere = hemisphere
            make_surfaces.inputs.maximum = 3.5
            make_surfaces.inputs.longitudinal = True
            make_surfaces.inputs.copy_inputs = True

            hemi_wf.connect([(copy_template_orig_white, make_surfaces,
                              [('out_file', 'orig_white')]),
                             (copy_template_white, make_surfaces,
                              [('out_file', 'in_orig')])])

        else:
            # If running single session
            # Tessellate by hemisphere
            """
            This is the step where the orig surface (ie, surf/?h.orig.nofix) is created.
            The surface is created by covering the filled hemisphere with triangles.
            Runs mri_pretess to create a connected WM volume (neighboring voxels must
            have faces in common) and then mri_tessellate to create the surface. The
            places where the points of the triangles meet are called vertices. Creates
            the file surf/?h.orig.nofix Note: the topology fixer will create the surface
            ?h.orig. Finally mris_extract_main_component will remove small surface
            components, not connected to the main body.
            """
            pretess2 = pe.Node(MRIPretess(), name='Pretess2')
            pretess2.inputs.out_file = 'filled-pretess{0}.mgz'.format(label)
            pretess2.inputs.label = label

            hemi_wf.connect([(hemi_inputspec, pretess2,
                              [('norm', 'in_norm'), ('filled', 'in_filled')])])

            tesselate = pe.Node(MRITessellate(), name="Tesselation")
            tesselate.inputs.out_file = "{0}.orig.nofix".format(hemisphere)
            tesselate.inputs.label_value = label
            hemi_wf.connect([(pretess2, tesselate, [('out_file', 'in_file')])])

            extract_main_component = pe.Node(
                ExtractMainComponent(), name="Extract_Main_Component")
            extract_main_component.inputs.out_file = "{0}.orig.nofix".format(
                hemisphere)
            hemi_wf.connect([(tesselate, extract_main_component,
                              [('surface', 'in_file')])])

            copy_orig = pe.Node(
                Function(['in_file', 'out_file'], ['out_file'], copy_file),
                name='Copy_Orig')
            copy_orig.inputs.out_file = '{0}.orig'.format(hemisphere)
            hemi_wf.connect([(extract_main_component, copy_orig,
                              [('out_file', 'in_file')])])

            # Orig Surface Smoothing 1
            """
            After tesselation, the orig surface is very jagged because each triangle is
            on the edge of a voxel face and so are at right angles to each other. The
            vertex positions are adjusted slightly here to reduce the angle. This is
            only necessary for the inflation processes. Creates surf/?h.smoothwm(.nofix).
            Calls mris_smooth. Smooth1 is the step just after tessellation.
            """

            smooth1 = pe.Node(SmoothTessellation(), name="Smooth1")
            smooth1.inputs.disable_estimates = True
            smooth1.inputs.seed = 1234
            smooth1.inputs.out_file = '{0}.smoothwm.nofix'.format(hemisphere)
            hemi_wf.connect([(extract_main_component, smooth1, [('out_file',
                                                                 'in_file')])])

            # Inflation 1
            """
            Inflation of the surf/?h.smoothwm(.nofix) surface to create surf/?h.inflated.
            The inflation attempts to minimize metric distortion so that distances and
            areas are preserved (ie, the surface is not stretched). In this sense, it is
            like inflating a paper bag and not a balloon. Inflate1 is the step just after
            tessellation.
            """

            inflate1 = pe.Node(MRIsInflate(), name="inflate1")
            inflate1.inputs.no_save_sulc = True
            inflate1.inputs.out_file = '{0}.inflated.nofix'.format(hemisphere)

            copy_inflate1 = pe.Node(
                Function(['in_file', 'out_file'], ['out_file'], copy_file),
                name='Copy_Inflate1')
            copy_inflate1.inputs.out_file = '{0}.inflated'.format(hemisphere)
            hemi_wf.connect([
                (smooth1, inflate1, [('surface', 'in_file')]),
                (inflate1, copy_inflate1, [('out_file', 'in_file')]),
            ])

            # Sphere
            """
            This is the initial step of automatic topology fixing. It is a
            quasi-homeomorphic spherical transformation of the inflated surface designed
            to localize topological defects for the subsequent automatic topology fixer.
            Calls mris_sphere.
            """

            qsphere = pe.Node(Sphere(), name="Sphere")
            qsphere.inputs.seed = 1234
            qsphere.inputs.magic = True
            qsphere.inputs.out_file = '{0}.qsphere.nofix'.format(hemisphere)
            if plugin_args:
                qsphere.plugin_args = plugin_args
            hemi_wf.connect([(inflate1, qsphere, [('out_file', 'in_file')]),
                             (hemi_inputspec, qsphere, [('num_threads',
                                                         'num_threads')])])

            # Automatic Topology Fixer
            """
            Finds topological defects (ie, holes in a filled hemisphere) using
            surf/?h.qsphere.nofix, and changes the orig surface (surf/?h.orig.nofix) to
            remove the defects. Changes the number of vertices. All the defects will be
            removed, but the user should check the orig surface in the volume to make
            sure that it looks appropriate.

            This mris_fix_topology does not take in the {lh,rh}.orig file, but instead takes in the
            subject ID and hemisphere and tries to find it from the subjects
            directory.
            """
            fix_topology = pe.Node(FixTopology(), name="Fix_Topology")
            fix_topology.inputs.mgz = True
            fix_topology.inputs.ga = True
            fix_topology.inputs.seed = 1234
            fix_topology.inputs.hemisphere = hemisphere
            fix_topology.inputs.copy_inputs = True
            hemi_wf.connect([(copy_orig, fix_topology,
                              [('out_file',
                                'in_orig')]), (copy_inflate1, fix_topology,
                                               [('out_file', 'in_inflated')]),
                             (qsphere, fix_topology, [('out_file', 'sphere')]),
                             (hemi_inputspec, fix_topology,
                              [('wm', 'in_wm'), ('brain', 'in_brain')])])

            # TODO: halt workflow for bad euler number
            euler_number = pe.Node(EulerNumber(), name="Euler_Number")

            hemi_wf.connect([
                (fix_topology, euler_number, [('out_file', 'in_file')]),
            ])

            remove_intersection = pe.Node(
                RemoveIntersection(), name="Remove_Intersection")
            remove_intersection.inputs.out_file = "{0}.orig".format(hemisphere)

            hemi_wf.connect([(euler_number, remove_intersection,
                              [('out_file', 'in_file')])])

            # White

            # This function implicitly calls other inputs based on the subject_id
            # need to make sure files are data sinked to the correct folders before
            # calling
            make_surfaces = pe.Node(MakeSurfaces(), name="Make_Surfaces")
            make_surfaces.inputs.noaparc = True
            make_surfaces.inputs.mgz = True
            make_surfaces.inputs.white_only = True
            make_surfaces.inputs.hemisphere = hemisphere
            make_surfaces.inputs.copy_inputs = True
            hemi_wf.connect([(remove_intersection, make_surfaces,
                              [('out_file', 'in_orig')]),
                             (hemi_inputspec, make_surfaces,
                              [('aseg', 'in_aseg'), ('t1', 'in_T1'),
                               ('filled', 'in_filled'), ('wm', 'in_wm')])])
            # end of non-longitudinal specific steps

        # Orig Surface Smoothing 2
        """
        After tesselation, the orig surface is very jagged because each triangle is on
        the edge of a voxel face and so are at right angles to each other. The vertex
        positions are adjusted slightly here to reduce the angle. This is only necessary
        for the inflation processes. Smooth2 is the step just after topology
        fixing.
        """
        smooth2 = pe.Node(SmoothTessellation(), name="Smooth2")
        smooth2.inputs.disable_estimates = True
        smooth2.inputs.smoothing_iterations = 3
        smooth2.inputs.seed = 1234
        smooth2.inputs.out_file = '{0}.smoothwm'.format(hemisphere)
        hemi_wf.connect([(make_surfaces, smooth2, [('out_white', 'in_file')])])

        # Inflation 2
        """
        Inflation of the surf/?h.smoothwm(.nofix) surface to create surf/?h.inflated.
        The inflation attempts to minimize metric distortion so that distances and areas
        are preserved (ie, the surface is not stretched). In this sense, it is like
        inflating a paper bag and not a balloon. Inflate2 is the step just after
        topology fixing.
        """
        inflate2 = pe.Node(MRIsInflate(), name="inflate2")
        inflate2.inputs.out_sulc = '{0}.sulc'.format(hemisphere)
        inflate2.inputs.out_file = '{0}.inflated'.format(hemisphere)
        hemi_wf.connect([
            (smooth2, inflate2, [('surface', 'in_file')]),
        ])

        # Compute Curvature
        """No documentation on this step"""

        curvature1 = pe.Node(Curvature(), name="Curvature1")
        curvature1.inputs.save = True
        curvature1.inputs.copy_input = True
        hemi_wf.connect([
            (make_surfaces, curvature1, [('out_white', 'in_file')]),
        ])

        curvature2 = pe.Node(Curvature(), name="Curvature2")
        curvature2.inputs.threshold = .999
        curvature2.inputs.n = True
        curvature2.inputs.averages = 5
        curvature2.inputs.save = True
        curvature2.inputs.distances = (10, 10)
        curvature1.inputs.copy_input = True
        hemi_wf.connect([
            (inflate2, curvature2, [('out_file', 'in_file')]),
        ])

        curvature_stats = pe.Node(CurvatureStats(), name="Curvature_Stats")
        curvature_stats.inputs.min_max = True
        curvature_stats.inputs.write = True
        curvature_stats.inputs.values = True
        curvature_stats.inputs.hemisphere = hemisphere
        curvature_stats.inputs.copy_inputs = True
        curvature_stats.inputs.out_file = '{0}.curv.stats'.format(hemisphere)
        hemi_wf.connect([
            (smooth2, curvature_stats, [('surface', 'surface')]),
            (make_surfaces, curvature_stats, [('out_curv', 'curvfile1')]),
            (inflate2, curvature_stats, [('out_sulc', 'curvfile2')]),
        ])

        if longitudinal:
            ar2_wf.connect([(inputspec, hemi_wf,
                             [('template_{0}_white'.format(hemisphere),
                               'Copy_Template_White.in_file'),
                              ('template_{0}_white'.format(hemisphere),
                               'Copy_Template_Orig_White.in_file'),
                              ('template_{0}_pial'.format(hemisphere),
                               'Copy_Template_Pial.in_file')])])

        # Connect inputs for the hemisphere workflows
        ar2_wf.connect(
            [(ca_normalize, hemi_wf,
              [('out_file', 'inputspec.norm')]), (fill, hemi_wf, [
                  ('out_file', 'inputspec.filled')
              ]), (copy_cc, hemi_wf, [('out_file', 'inputspec.aseg')]),
             (mri_mask, hemi_wf, [('out_file', 'inputspec.t1')]),
             (pretess, hemi_wf, [('out_file',
                                  'inputspec.wm')]), (normalization2, hemi_wf,
                                                      [('out_file',
                                                        'inputspec.brain')]),
             (inputspec, hemi_wf, [('num_threads', 'inputspec.num_threads')])])

        # Outputs for hemisphere workflow
        hemi_outputs = [
            'orig_nofix', 'orig', 'smoothwm_nofix', 'inflated_nofix',
            'qsphere_nofix', 'white', 'curv', 'area', 'cortex', 'pial_auto',
            'thickness', 'smoothwm', 'sulc', 'inflated', 'white_H', 'white_K',
            'inflated_H', 'inflated_K', 'curv_stats'
        ]

        hemi_outputspec = pe.Node(
            IdentityInterface(fields=hemi_outputs), name="outputspec")

        hemi_wf.connect(
            [(extract_main_component, hemi_outputspec,
              [('out_file', 'orig_nofix')]), (inflate1, hemi_outputspec, [
                  ('out_file', 'inflated_nofix')
              ]), (smooth1, hemi_outputspec, [('surface', 'smoothwm_nofix')]),
             (qsphere, hemi_outputspec, [('out_file', 'qsphere_nofix')]),
             (remove_intersection, hemi_outputspec,
              [('out_file', 'orig')]), (make_surfaces, hemi_outputspec, [
                  ('out_white', 'white'), ('out_curv', 'curv'),
                  ('out_area', 'area'), ('out_cortex', 'cortex'), ('out_pial',
                                                                   'pial_auto')
              ]), (smooth2, hemi_outputspec,
                   [('surface', 'smoothwm')]), (inflate2, hemi_outputspec,
                                                [('out_sulc', 'sulc'),
                                                 ('out_file', 'inflated')]),
             (curvature1, hemi_outputspec,
              [('out_mean', 'white_H'),
               ('out_gauss', 'white_K')]), (curvature2, hemi_outputspec, [
                   ('out_mean', 'inflated_H'), ('out_gauss', 'inflated_K')
               ]), (curvature_stats, hemi_outputspec, [('out_file',
                                                        'curv_stats')])])

    outputs = [
        'nu', 'tal_lta', 'norm', 'ctrl_pts', 'tal_m3z', 'nu_noneck',
        'talskull2', 'aseg_noCC', 'cc_up', 'aseg_auto', 'aseg_presurf',
        'brain', 'brain_finalsurfs', 'wm_seg', 'wm_aseg', 'wm', 'ponscc_log',
        'filled'
    ]
    for hemi in ('lh', 'rh'):
        for field in hemi_outputs:
            outputs.append("{0}_".format(hemi) + field)
    outputspec = pe.Node(IdentityInterface(fields=outputs), name="outputspec")

    if fsvernum >= 6:
        ar2_wf.connect([(add_to_header_nu, outputspec, [('out_file', 'nu')])])
    else:
        # add to outputspec to perserve datasinking
        ar2_wf.connect([(inputspec, outputspec, [('nu', 'nu')])])

    ar2_wf.connect([
        (align_transform, outputspec, [('out_file', 'tal_lta')]),
        (ca_normalize, outputspec, [('out_file', 'norm')]),
        (ca_normalize, outputspec, [('control_points', 'ctrl_pts')]),
        (ca_register, outputspec, [('out_file', 'tal_m3z')]),
        (remove_neck, outputspec, [('out_file', 'nu_noneck')]),
        (em_reg_withskull, outputspec, [('out_file', 'talskull2')]),
        (ca_label, outputspec, [('out_file', 'aseg_noCC')]),
        (segment_cc, outputspec, [('out_rotation', 'cc_up'), ('out_file',
                                                              'aseg_auto')]),
        (copy_cc, outputspec, [('out_file', 'aseg_presurf')]),
        (normalization2, outputspec, [('out_file', 'brain')]),
        (mri_mask, outputspec, [('out_file', 'brain_finalsurfs')]),
        (wm_seg, outputspec, [('out_file', 'wm_seg')]),
        (edit_wm, outputspec, [('out_file', 'wm_aseg')]),
        (pretess, outputspec, [('out_file', 'wm')]),
        (fill, outputspec, [('out_file', 'filled'), ('log_file',
                                                     'ponscc_log')]),
    ])

    for hemi, hemi_wf in [('lh', ar2_lh), ('rh', ar2_rh)]:
        for field in hemi_outputs:
            output = "{0}_".format(hemi) + field
            ar2_wf.connect([(hemi_wf, outputspec, [("outputspec." + field,
                                                    output)])])

    return ar2_wf, outputs



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import Denoise


def test_Denoise_inputs():
    input_map = dict(
        block_radius=dict(usedefault=True, ),
        in_file=dict(mandatory=True, ),
        in_mask=dict(),
        noise_mask=dict(),
        noise_model=dict(
            mandatory=True,
            usedefault=True,
        ),
        patch_radius=dict(usedefault=True, ),
        signal_mask=dict(),
        snr=dict(),
    )
    inputs = Denoise.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Denoise_outputs():
    output_map = dict(out_file=dict(), )
    outputs = Denoise.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from future import standard_library
standard_library.install_aliases()

from ..nipype2boutiques import generate_boutiques_descriptor


def test_generate():
    generate_boutiques_descriptor(module='nipype.interfaces.ants.registration',
                                  interface_name='ANTS',
                                  ignored_template_inputs=(),
                                  docker_image=None,
                                  docker_index=None,
                                  verbose=False,
                                  ignore_template_numbers=False)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import os
from nipype.interfaces.fsl.base import no_fsl, no_fsl_course_data
import nipype.pipeline.engine as pe
import nipype.interfaces.utility as util
import pytest
import tempfile
import shutil
from subprocess import call
from nipype.workflows.dmri.fsl.tbss import create_tbss_all
import nipype.interfaces.io as nio
from nipype.interfaces import fsl


def _tbss_test_helper(estimate_skeleton):
    fsl_course_dir = os.path.abspath(os.environ['FSL_COURSE_DATA'])
    fsl.FSLCommand.set_default_output_type('NIFTI_GZ')
    test_dir = tempfile.mkdtemp(prefix="nipype_test_tbss_")
    tbss_orig_dir = os.path.join(test_dir, "tbss_all_original")
    os.mkdir(tbss_orig_dir)
    old_dir = os.getcwd()
    os.chdir(tbss_orig_dir)

    subjects = ['1260', '1549']
    FA_list = [
        os.path.join(fsl_course_dir, 'tbss', subject_id + '.nii.gz')
        for subject_id in subjects
    ]
    for f in FA_list:
        shutil.copy(f, os.getcwd())

    call(
        ['tbss_1_preproc'] +
        [subject_id + '.nii.gz' for subject_id in subjects],
        env=os.environ.update({
            'FSLOUTPUTTYPE': 'NIFTI_GZ'
        }))
    tbss1_orig_dir = os.path.join(test_dir, "tbss1_original")
    shutil.copytree(tbss_orig_dir, tbss1_orig_dir)

    call(
        ['tbss_2_reg', '-T'],
        env=os.environ.update({
            'FSLOUTPUTTYPE': 'NIFTI_GZ'
        }))
    tbss2_orig_dir = os.path.join(test_dir, "tbss2_original")
    shutil.copytree(tbss_orig_dir, tbss2_orig_dir)

    if estimate_skeleton:
        call(
            ['tbss_3_postreg', '-S'],
            env=os.environ.update({
                'FSLOUTPUTTYPE': 'NIFTI_GZ'
            }))
    else:
        call(
            ['tbss_3_postreg', '-T'],
            env=os.environ.update({
                'FSLOUTPUTTYPE': 'NIFTI_GZ'
            }))
    tbss3_orig_dir = os.path.join(test_dir, "tbss3_original")
    shutil.copytree(tbss_orig_dir, tbss3_orig_dir)

    call(
        ['tbss_4_prestats', '0.2'],
        env=os.environ.update({
            'FSLOUTPUTTYPE': 'NIFTI_GZ'
        }))
    tbss4_orig_dir = os.path.join(test_dir, "tbss4_original")
    shutil.copytree(tbss_orig_dir, tbss4_orig_dir)

    pipeline = pe.Workflow(name="test_tbss")
    pipeline.base_dir = os.path.join(test_dir, "tbss_nipype")

    tbss = create_tbss_all(estimate_skeleton=estimate_skeleton)
    tbss.inputs.inputnode.fa_list = FA_list
    tbss.inputs.inputnode.skeleton_thresh = 0.2

    tbss1_original_datasource = pe.Node(
        nio.DataGrabber(
            outfields=['fa_list', 'mask_list'], sort_filelist=False),
        name='tbss1_original_datasource')
    tbss1_original_datasource.inputs.base_directory = tbss1_orig_dir
    tbss1_original_datasource.inputs.template = 'FA/%s_FA%s.nii.gz'
    tbss1_original_datasource.inputs.template_args = dict(
        fa_list=[[subjects, '']], mask_list=[[subjects, '_mask']])

    tbss1_test_fa = pe.MapNode(
        util.AssertEqual(),
        name="tbss1_fa_test",
        iterfield=['volume1', 'volume2'])
    tbss1_test_mask = pe.MapNode(
        util.AssertEqual(),
        name="tbss1_mask_test",
        iterfield=['volume1', 'volume2'])

    pipeline.connect(tbss, 'tbss1.outputnode.fa_list', tbss1_test_fa,
                     'volume1')
    pipeline.connect(tbss, 'tbss1.outputnode.mask_list', tbss1_test_mask,
                     'volume1')
    pipeline.connect(tbss1_original_datasource, 'fa_list', tbss1_test_fa,
                     'volume2')
    pipeline.connect(tbss1_original_datasource, 'mask_list', tbss1_test_mask,
                     'volume2')
    tbss2_original_datasource = pe.Node(
        nio.DataGrabber(outfields=['field_list'], sort_filelist=False),
        name='tbss2_original_datasource')

    tbss2_original_datasource.inputs.base_directory = tbss2_orig_dir
    tbss2_original_datasource.inputs.template = 'FA/%s_FA%s.nii.gz'
    tbss2_original_datasource.inputs.template_args = dict(
        field_list=[[subjects, '_to_target_warp']])
    tbss2_test_field = pe.MapNode(
        util.AssertEqual(),
        name="tbss2_test_field",
        iterfield=['volume1', 'volume2'])

    pipeline.connect(tbss, 'tbss2.outputnode.field_list', tbss2_test_field,
                     'volume1')
    pipeline.connect(tbss2_original_datasource, 'field_list', tbss2_test_field,
                     'volume2')

    tbss3_original_datasource = pe.Node(
        nio.DataGrabber(
            outfields=[
                'groupmask', 'skeleton_file', 'meanfa_file', 'mergefa_file'
            ],
            sort_filelist=False),
        name='tbss3_original_datasource')
    tbss3_original_datasource.inputs.base_directory = tbss3_orig_dir
    tbss3_original_datasource.inputs.template = 'stats/%s.nii.gz'
    tbss3_original_datasource.inputs.template_args = dict(
        groupmask=[['mean_FA_mask']],
        skeleton_file=[['mean_FA_skeleton']],
        meanfa_file=[['mean_FA']],
        mergefa_file=[['all_FA']])

    tbss3_test_groupmask = pe.Node(
        util.AssertEqual(), name="tbss3_test_groupmask")
    tbss3_test_skeleton_file = pe.Node(
        util.AssertEqual(), name="tbss3_test_skeleton_file")
    tbss3_test_meanfa_file = pe.Node(
        util.AssertEqual(), name="tbss3_test_meanfa_file")
    tbss3_test_mergefa_file = pe.Node(
        util.AssertEqual(), name="tbss3_test_mergefa_file")

    pipeline.connect(tbss, 'tbss3.outputnode.groupmask', tbss3_test_groupmask,
                     'volume1')
    pipeline.connect(tbss3_original_datasource, 'groupmask',
                     tbss3_test_groupmask, 'volume2')
    pipeline.connect(tbss, 'tbss3.outputnode.skeleton_file',
                     tbss3_test_skeleton_file, 'volume1')
    pipeline.connect(tbss3_original_datasource, 'skeleton_file',
                     tbss3_test_skeleton_file, 'volume2')
    pipeline.connect(tbss, 'tbss3.outputnode.meanfa_file',
                     tbss3_test_meanfa_file, 'volume1')
    pipeline.connect(tbss3_original_datasource, 'meanfa_file',
                     tbss3_test_meanfa_file, 'volume2')
    pipeline.connect(tbss, 'tbss3.outputnode.mergefa_file',
                     tbss3_test_mergefa_file, 'volume1')
    pipeline.connect(tbss3_original_datasource, 'mergefa_file',
                     tbss3_test_mergefa_file, 'volume2')

    tbss4_original_datasource = pe.Node(
        nio.DataGrabber(
            outfields=['all_FA_skeletonised', 'mean_FA_skeleton_mask'],
            sort_filelist=False),
        name='tbss4_original_datasource')
    tbss4_original_datasource.inputs.base_directory = tbss4_orig_dir
    tbss4_original_datasource.inputs.template = 'stats/%s.nii.gz'
    tbss4_original_datasource.inputs.template_args = dict(
        all_FA_skeletonised=[['all_FA_skeletonised']],
        mean_FA_skeleton_mask=[['mean_FA_skeleton_mask']])
    tbss4_test_all_FA_skeletonised = pe.Node(
        util.AssertEqual(), name="tbss4_test_all_FA_skeletonised")
    tbss4_test_mean_FA_skeleton_mask = pe.Node(
        util.AssertEqual(), name="tbss4_test_mean_FA_skeleton_mask")

    pipeline.connect(tbss, 'tbss4.outputnode.projectedfa_file',
                     tbss4_test_all_FA_skeletonised, 'volume1')
    pipeline.connect(tbss4_original_datasource, 'all_FA_skeletonised',
                     tbss4_test_all_FA_skeletonised, 'volume2')
    pipeline.connect(tbss, 'tbss4.outputnode.skeleton_mask',
                     tbss4_test_mean_FA_skeleton_mask, 'volume1')
    pipeline.connect(tbss4_original_datasource, 'mean_FA_skeleton_mask',
                     tbss4_test_mean_FA_skeleton_mask, 'volume2')

    pipeline.run(plugin='Linear')
    os.chdir(old_dir)
    shutil.rmtree(test_dir)


# this test is disabled until we figure out what is wrong with TBSS in 5.0.9


@pytest.mark.skipif(no_fsl(), reason="fsl is not installed")
@pytest.mark.skipif(no_fsl_course_data(), reason="fsl data not available")
def test_disabled_tbss_est_skeleton():
    _tbss_test_helper(True)


# this test is disabled until we figure out what is wrong with TBSS in 5.0.9


@pytest.mark.skipif(no_fsl(), reason="fsl is not installed")
@pytest.mark.skipif(no_fsl_course_data(), reason="fsl data not available")
def test_disabled_tbss_est_skeleton_use_precomputed_skeleton():
    _tbss_test_helper(False)
